import pytest
from thinc.api import Ragged, get_current_ops
import spacy

from curated_transformers.tokenization.char_encoder import (
    build_char_encoder_loader_v1,
    build_char_encoder_v1,
)
from curated_transformers.tokenization.hf_loader import build_hf_piece_encoder_loader_v1
from curated_transformers.tokenization.wordpiece_encoder import (
    build_wordpiece_encoder_v1,
)
from curated_transformers.util import registry
from curated_transformers._compat import has_fugashi, has_hf_transformers, has_sudachi


def test_char_encoder(test_dir):
    ops = get_current_ops()
    encoder = build_char_encoder_v1()
    encoder.init = build_char_encoder_loader_v1(path=test_dir / "toy-chars.txt")
    encoder.initialize()

    nlp = spacy.blank("nl")
    sample_docs = [nlp.make_doc("Zeeën van tijd."), nlp.make_doc("Geïnd geld")]

    encoding = encoder.predict(sample_docs)

    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops.xp.testing.assert_array_equal(encoding[0].lengths, [1, 5, 3, 4, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd, [2, 56, 9, 9, 57, 18, 26, 5, 18, 24, 13, 14, 8, 1, 3]
    )

    assert isinstance(encoding[1], Ragged)
    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 5, 4, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd, [2, 37, 9, 1, 18, 8, 11, 9, 16, 8, 3]
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.skipif(not has_fugashi, reason="requires fugashi")
@pytest.mark.skipif(not has_sudachi, reason="requires SudachiPy and SudachiDict-core")
def test_char_encoder_hf_model():
    ops = get_current_ops()
    encoder = build_char_encoder_v1()
    encoder.init = registry.model_loaders.get(
        "curated-transformers.HFPieceEncoderLoader.v1"
    )(name="cl-tohoku/bert-base-japanese-char-v2")
    encoder.initialize()

    nlp = spacy.blank("ja")
    sample_docs = [nlp.make_doc("日本語だよ"), nlp.make_doc("吾輩は猫である。")]

    encoding = encoder.predict(sample_docs)

    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops.xp.testing.assert_array_equal(encoding[0].lengths, [1, 2, 1, 1, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd, [2, 2719, 2828, 4923, 882, 922, 3]
    )

    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 2, 1, 1, 1, 2, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd, [2, 1583, 5159, 897, 3574, 889, 852, 925, 829, 3]
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.skipif(not has_fugashi, reason="requires fugashi")
@pytest.mark.skipif(not has_sudachi, reason="requires SudachiPy and SudachiDict-core")
def test_hf_loader_rejects_incorrect_encoder():
    encoder = build_wordpiece_encoder_v1()
    encoder.init = build_hf_piece_encoder_loader_v1(
        name="cl-tohoku/bert-base-japanese-char-v2"
    )
    with pytest.raises(ValueError, match="incompatible piece encoder"):
        encoder.initialize()


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.skipif(not has_fugashi, reason="requires fugashi")
@pytest.mark.skipif(not has_sudachi, reason="requires SudachiPy and SudachiDict-core")
def test_hf_loader_rejects_incorrect_model():
    encoder = build_char_encoder_v1()
    encoder.init = build_hf_piece_encoder_loader_v1(
        name="cl-tohoku/bert-base-japanese-v2"
    )
    with pytest.raises(ValueError, match="only support character subword"):
        encoder.initialize()


@pytest.mark.slow
def test_loader_rejects_incorrect_encoder(test_dir):
    encoder = build_wordpiece_encoder_v1()
    encoder.init = build_char_encoder_loader_v1(path=test_dir / "toy-chars.txt")
    with pytest.raises(ValueError, match="incompatible model"):
        encoder.initialize()



from typing import Generic, Optional, TypeVar
from torch import Tensor
from torch.nn import Module


from .albert.encoder import AlbertEncoder
from .bert.encoder import BertEncoder
from .roberta.encoder import RobertaEncoder

from .attention import AttentionMask
from ..output import PyTorchTransformerOutput

CuratedEncoderT = TypeVar("CuratedEncoderT", AlbertEncoder, BertEncoder, RobertaEncoder)


class CuratedTransformer(Generic[CuratedEncoderT], Module):
    """Simple wrapper for encoders. Currently only used to add a predictable
    prefix (curated_encoder) to encoders."""

    def __init__(self, encoder: CuratedEncoderT) -> None:
        super().__init__()

        # Type ignore, because TorchScript does not allow Module
        # as a class variable type.
        self.curated_encoder = encoder  # type: ignore[var-annotated]

    def forward(
        self,
        input_ids: Tensor,
        attention_mask: Optional[AttentionMask] = None,
        token_type_ids: Optional[Tensor] = None,
    ) -> PyTorchTransformerOutput:
        """
        Shapes:
            input_ids, attention_mask, token_type_ids - (batch, seq_len)
        """
        return self.curated_encoder.forward(input_ids, attention_mask, token_type_ids)



from typing import List, Optional, Tuple, Callable, Any, cast
from pathlib import Path
from functools import partial
from spacy.tokens import Doc
from spacy.util import SimpleFrozenDict
from thinc.api import (
    Model,
    PyTorchWrapper_v2,
    xp2torch,
    torch2xp,
    get_torch_default_device,
    TorchScriptWrapper_v1,
)
from thinc.layers import chain
from thinc.model import Model
from thinc.shims.pytorch_grad_scaler import PyTorchGradScaler
from thinc.types import ArgsKwargs, Floats2d, Ints1d
import torch


from .output import TransformerModelOutput, PyTorchTransformerOutput
from .pytorch.albert import AlbertConfig, AlbertEncoder
from .pytorch.bert import BertConfig, BertEncoder
from .pytorch.curated_transformer import CuratedTransformer, CuratedEncoderT
from .pytorch.hf_util import convert_pretrained_model_for_encoder
from .pytorch.roberta import RobertaConfig, RobertaEncoder
from .remove_eos_bos import remove_bos_eos
from .with_non_ws_tokens import with_non_ws_tokens
from ..tokenization.types import Tok2PiecesModelT
from .types import (
    TorchTransformerInT,
    TorchTransformerModelT,
    TorchTransformerOutT,
    TransformerBackpropT,
    TransformerInT,
    TransformerOutT,
    TransformerModelT,
    SpanExtractorModelT,
)
from ..errors import Errors


def build_albert_transformer_model_v1(
    *,
    vocab_size: int,
    with_spans: Callable[
        [TorchTransformerModelT],
        SpanExtractorModelT,
    ],
    piece_encoder: Tok2PiecesModelT,
    attention_probs_dropout_prob: float = 0.0,
    embedding_width: int = 128,
    hidden_act: str = "gelu_new",
    hidden_dropout_prob: float = 0.0,
    hidden_width: int = 768,
    intermediate_width: int = 3072,
    layer_norm_eps: float = 1e-12,
    max_position_embeddings: int = 512,
    model_max_length: int = 512,
    num_attention_heads: int = 12,
    num_hidden_groups: int = 1,
    num_hidden_layers: int = 12,
    padding_idx: int = 0,
    type_vocab_size: int = 2,
    torchscript: bool = False,
    mixed_precision: bool = False,
    grad_scaler_config: dict = SimpleFrozenDict(),
) -> TransformerModelT:
    """Construct an ALBERT transformer model.

    vocab_size (int):
        Vocabulary size.
    with_spans (Callable):
        Callback that constructs a span generator model.
    piece_encoder (Model)
        The piece encoder to segment input tokens.
    attention_probs_dropout_prob (float):
        Dropout probabilty of the self-attention layers.
    embedding_width (int):
        Width of the embedding representations.
    hidden_act (str):
        Activation used by the point-wise feed-forward layers.
    hidden_dropout_prob (float):
        Dropout probabilty of the point-wise feed-forward and
        embedding layers.
    hidden_width (int):
        Width of the final representations.
    intermediate_width (int):
        Width of the intermediate projection layer in the
        point-wise feed-forward layer.
    layer_norm_eps (float):
        Epsilon for layer normalization.
    max_position_embeddings (int):
        Maximum length of position embeddings.
    model_max_length (int):
        Maximum length of model inputs.
    num_attention_heads (int):
        Number of self-attention heads.
    num_hidden_groups (int):
        Number of layer groups whose constituents share parameters.
    num_hidden_layers (int):
        Number of hidden layers.
    padding_idx (int):
        Index of the padding meta-token.
    type_vocab_size (int):
        Type vocabulary size.
    torchscript (bool):
        Set to `True` when loading TorchScript models, `False` otherwise.
    mixed_precision (bool):
        Use mixed-precision training.
    grad_scaler_config (dict):
        Configuration passed to the PyTorch gradient scaler.
    """
    config = AlbertConfig(
        embedding_width=embedding_width,
        hidden_width=hidden_width,
        intermediate_width=intermediate_width,
        num_attention_heads=num_attention_heads,
        num_hidden_groups=num_hidden_groups,
        num_hidden_layers=num_hidden_layers,
        attention_probs_dropout_prob=attention_probs_dropout_prob,
        hidden_dropout_prob=hidden_dropout_prob,
        hidden_act=hidden_act,
        vocab_size=vocab_size,
        type_vocab_size=type_vocab_size,
        max_position_embeddings=max_position_embeddings,
        model_max_length=model_max_length,
        layer_norm_eps=layer_norm_eps,
        padding_idx=padding_idx,
    )

    if torchscript:
        transformer = _torchscript_encoder(
            model_max_length=model_max_length, padding_idx=padding_idx
        )
    else:
        encoder = AlbertEncoder(config)
        transformer = _pytorch_encoder(
            encoder,
            mixed_precision=mixed_precision,
            grad_scaler_config=grad_scaler_config,
        )

    return build_transformer_model_v1(
        with_spans=with_spans,
        piece_encoder=piece_encoder,
        transformer=transformer,
    )


def build_bert_transformer_model_v1(
    *,
    vocab_size: int,
    with_spans: Callable[
        [TorchTransformerModelT],
        SpanExtractorModelT,
    ],
    piece_encoder: Tok2PiecesModelT,
    attention_probs_dropout_prob: float = 0.1,
    hidden_act: str = "gelu",
    hidden_dropout_prob: float = 0.1,
    hidden_width: int = 768,
    intermediate_width: int = 3072,
    layer_norm_eps: float = 1e-12,
    max_position_embeddings: int = 512,
    model_max_length: int = 512,
    num_attention_heads: int = 12,
    num_hidden_layers: int = 12,
    padding_idx: int = 0,
    type_vocab_size: int = 2,
    torchscript: bool = False,
    mixed_precision: bool = False,
    grad_scaler_config: dict = SimpleFrozenDict(),
) -> TransformerModelT:
    """Construct a BERT transformer model.

    vocab_size (int):
        Vocabulary size.
    with_spans (Callable):
        Callback that constructs a span generator model.
    piece_encoder (Model)
        The piece encoder to segment input tokens.
    attention_probs_dropout_prob (float):
        Dropout probabilty of the self-attention layers.
    hidden_act (str):
        Activation used by the point-wise feed-forward layers.
    hidden_dropout_prob (float):
        Dropout probabilty of the point-wise feed-forward and
        embedding layers.
    hidden_width (int):
        Width of the final representations.
    intermediate_width (int):
        Width of the intermediate projection layer in the
        point-wise feed-forward layer.
    layer_norm_eps (float):
        Epsilon for layer normalization.
    max_position_embeddings (int):
        Maximum length of position embeddings.
    model_max_length (int):
        Maximum length of model inputs.
    num_attention_heads (int):
        Number of self-attention heads.
    num_hidden_layers (int):
        Number of hidden layers.
    padding_idx (int):
        Index of the padding meta-token.
    type_vocab_size (int):
        Type vocabulary size.
    torchscript (bool):
        Set to `True` when loading TorchScript models, `False` otherwise.
    mixed_precision (bool):
        Use mixed-precision training.
    grad_scaler_config (dict):
        Configuration passed to the PyTorch gradient scaler.
    """
    config = BertConfig(
        hidden_width=hidden_width,
        intermediate_width=intermediate_width,
        num_attention_heads=num_attention_heads,
        num_hidden_layers=num_hidden_layers,
        attention_probs_dropout_prob=attention_probs_dropout_prob,
        hidden_dropout_prob=hidden_dropout_prob,
        hidden_act=hidden_act,
        vocab_size=vocab_size,
        type_vocab_size=type_vocab_size,
        max_position_embeddings=max_position_embeddings,
        model_max_length=model_max_length,
        layer_norm_eps=layer_norm_eps,
        padding_idx=padding_idx,
    )

    if torchscript:
        transformer = _torchscript_encoder(
            model_max_length=model_max_length, padding_idx=padding_idx
        )
    else:
        encoder = BertEncoder(config)
        transformer = _pytorch_encoder(
            encoder,
            mixed_precision=mixed_precision,
            grad_scaler_config=grad_scaler_config,
        )

    return build_transformer_model_v1(
        with_spans=with_spans,
        piece_encoder=piece_encoder,
        transformer=transformer,
    )


def build_camembert_transformer_model_v1(
    *,
    vocab_size: int,
    with_spans: Callable[
        [TorchTransformerModelT],
        SpanExtractorModelT,
    ],
    piece_encoder: Tok2PiecesModelT,
    attention_probs_dropout_prob: float = 0.1,
    hidden_act: str = "gelu",
    hidden_dropout_prob: float = 0.1,
    hidden_width: int = 768,
    intermediate_width: int = 3072,
    layer_norm_eps: float = 1e-5,
    max_position_embeddings: int = 514,
    model_max_length: int = 512,
    num_attention_heads: int = 12,
    num_hidden_layers: int = 12,
    padding_idx: int = 1,
    type_vocab_size: int = 1,
    torchscript=False,
) -> TransformerModelT:
    """Construct a CamemBERT transformer model.

    vocab_size (int):
        Vocabulary size.
    with_spans (Callable):
        Callback that constructs a span generator model.
    piece_encoder (Model)
        The piece encoder to segment input tokens.
    attention_probs_dropout_prob (float):
        Dropout probabilty of the self-attention layers.
    hidden_act (str):
        Activation used by the point-wise feed-forward layers.
    hidden_dropout_prob (float):
        Dropout probabilty of the point-wise feed-forward and
        embedding layers.
    hidden_width (int):
        Width of the final representations.
    intermediate_width (int):
        Width of the intermediate projection layer in the
        point-wise feed-forward layer.
    layer_norm_eps (float):
        Epsilon for layer normalization.
    max_position_embeddings (int):
        Maximum length of position embeddings.
    model_max_length (int):
        Maximum length of model inputs.
    num_attention_heads (int):
        Number of self-attention heads.
    num_hidden_layers (int):
        Number of hidden layers.
    padding_idx (int):
        Index of the padding meta-token.
    type_vocab_size (int):
        Type vocabulary size.
    torchscript (bool):
        Set to `True` when loading TorchScript models, `False` otherwise.
    mixed_precision (bool):
        Use mixed-precision training.
    grad_scaler_config (dict):
        Configuration passed to the PyTorch gradient scaler.
    """
    config = RobertaConfig(
        hidden_width=hidden_width,
        intermediate_width=intermediate_width,
        num_attention_heads=num_attention_heads,
        num_hidden_layers=num_hidden_layers,
        attention_probs_dropout_prob=attention_probs_dropout_prob,
        hidden_dropout_prob=hidden_dropout_prob,
        hidden_act=hidden_act,
        vocab_size=vocab_size,
        type_vocab_size=type_vocab_size,
        max_position_embeddings=max_position_embeddings,
        model_max_length=model_max_length,
        layer_norm_eps=layer_norm_eps,
        padding_idx=padding_idx,
    )

    if torchscript:
        transformer = _torchscript_encoder(
            model_max_length=model_max_length, padding_idx=padding_idx
        )
    else:
        encoder = RobertaEncoder(config)
        transformer = _pytorch_encoder(encoder)

    return build_transformer_model_v1(
        with_spans=with_spans,
        piece_encoder=piece_encoder,
        transformer=transformer,
    )


def build_roberta_transformer_model_v1(
    *,
    vocab_size: int,
    with_spans: Callable[
        [TorchTransformerModelT],
        SpanExtractorModelT,
    ],
    piece_encoder: Tok2PiecesModelT,
    attention_probs_dropout_prob: float = 0.1,
    hidden_act: str = "gelu",
    hidden_dropout_prob: float = 0.1,
    hidden_width: int = 768,
    intermediate_width: int = 3072,
    layer_norm_eps: float = 1e-5,
    max_position_embeddings: int = 514,
    model_max_length: int = 512,
    num_attention_heads: int = 12,
    num_hidden_layers: int = 12,
    padding_idx: int = 1,
    type_vocab_size: int = 1,
    torchscript: bool = False,
    mixed_precision: bool = False,
    grad_scaler_config: dict = SimpleFrozenDict(),
) -> TransformerModelT:
    """Construct a RoBERTa transformer model.

    vocab_size (int):
        Vocabulary size.
    with_spans (Callable):
        Callback that constructs a span generator model.
    piece_encoder (Model)
        The piece encoder to segment input tokens.
    attention_probs_dropout_prob (float):
        Dropout probabilty of the self-attention layers.
    hidden_act (str):
        Activation used by the point-wise feed-forward layers.
    hidden_dropout_prob (float):
        Dropout probabilty of the point-wise feed-forward and
        embedding layers.
    hidden_width (int):
        Width of the final representations.
    intermediate_width (int):
        Width of the intermediate projection layer in the
        point-wise feed-forward layer.
    layer_norm_eps (float):
        Epsilon for layer normalization.
    max_position_embeddings (int):
        Maximum length of position embeddings.
    model_max_length (int):
        Maximum length of model inputs.
    num_attention_heads (int):
        Number of self-attention heads.
    num_hidden_layers (int):
        Number of hidden layers.
    padding_idx (int):
        Index of the padding meta-token.
    type_vocab_size (int):
        Type vocabulary size.
    torchscript (bool):
        Set to `True` when loading TorchScript models, `False` otherwise.
    mixed_precision (bool):
        Use mixed-precision training.
    grad_scaler_config (dict):
        Configuration passed to the PyTorch gradient scaler.
    """
    config = RobertaConfig(
        hidden_width=hidden_width,
        intermediate_width=intermediate_width,
        num_attention_heads=num_attention_heads,
        num_hidden_layers=num_hidden_layers,
        attention_probs_dropout_prob=attention_probs_dropout_prob,
        hidden_dropout_prob=hidden_dropout_prob,
        hidden_act=hidden_act,
        vocab_size=vocab_size,
        type_vocab_size=type_vocab_size,
        max_position_embeddings=max_position_embeddings,
        model_max_length=model_max_length,
        layer_norm_eps=layer_norm_eps,
        padding_idx=padding_idx,
    )

    if torchscript:
        transformer = _torchscript_encoder(
            model_max_length=model_max_length, padding_idx=padding_idx
        )
    else:
        encoder = RobertaEncoder(config)
        transformer = _pytorch_encoder(
            encoder,
            mixed_precision=mixed_precision,
            grad_scaler_config=grad_scaler_config,
        )

    return build_transformer_model_v1(
        with_spans=with_spans,
        piece_encoder=piece_encoder,
        transformer=transformer,
    )


def build_xlmr_transformer_model_v1(
    *,
    vocab_size: int,
    with_spans: Callable[
        [TorchTransformerModelT],
        SpanExtractorModelT,
    ],
    piece_encoder: Tok2PiecesModelT,
    attention_probs_dropout_prob: float = 0.1,
    hidden_act: str = "gelu",
    hidden_dropout_prob: float = 0.1,
    hidden_width: int = 768,
    intermediate_width: int = 3072,
    layer_norm_eps: float = 1e-5,
    max_position_embeddings: int = 514,
    model_max_length: int = 512,
    num_attention_heads: int = 12,
    num_hidden_layers: int = 12,
    padding_idx: int = 1,
    type_vocab_size: int = 1,
    torchscript: bool = False,
    mixed_precision: bool = False,
    grad_scaler_config: dict = SimpleFrozenDict(),
) -> TransformerModelT:
    """Construct a XLM-RoBERTa transformer model.

    vocab_size (int):
        Vocabulary size.
    with_spans (Callable):
        Callback that constructs a span generator model.
    piece_encoder (Model)
        The piece encoder to segment input tokens.
    attention_probs_dropout_prob (float):
        Dropout probabilty of the self-attention layers.
    hidden_act (str):
        Activation used by the point-wise feed-forward layers.
    hidden_dropout_prob (float):
        Dropout probabilty of the point-wise feed-forward and
        embedding layers.
    hidden_width (int):
        Width of the final representations.
    intermediate_width (int):
        Width of the intermediate projection layer in the
        point-wise feed-forward layer.
    layer_norm_eps (float):
        Epsilon for layer normalization.
    max_position_embeddings (int):
        Maximum length of position embeddings.
    model_max_length (int):
        Maximum length of model inputs.
    num_attention_heads (int):
        Number of self-attention heads.
    num_hidden_layers (int):
        Number of hidden layers.
    padding_idx (int):
        Index of the padding meta-token.
    type_vocab_size (int):
        Type vocabulary size.
    torchscript (bool):
        Set to `True` when loading TorchScript models, `False` otherwise.
    mixed_precision (bool):
        Use mixed-precision training.
    grad_scaler_config (dict):
        Configuration passed to the PyTorch gradient scaler.
    """
    config = RobertaConfig(
        hidden_width=hidden_width,
        intermediate_width=intermediate_width,
        num_attention_heads=num_attention_heads,
        num_hidden_layers=num_hidden_layers,
        attention_probs_dropout_prob=attention_probs_dropout_prob,
        hidden_dropout_prob=hidden_dropout_prob,
        hidden_act=hidden_act,
        vocab_size=vocab_size,
        type_vocab_size=type_vocab_size,
        max_position_embeddings=max_position_embeddings,
        model_max_length=model_max_length,
        layer_norm_eps=layer_norm_eps,
        padding_idx=padding_idx,
    )

    if torchscript:
        transformer = _torchscript_encoder(
            model_max_length=model_max_length, padding_idx=padding_idx
        )
    else:
        encoder = RobertaEncoder(config)
        transformer = _pytorch_encoder(
            encoder,
            mixed_precision=mixed_precision,
            grad_scaler_config=grad_scaler_config,
        )

    return build_transformer_model_v1(
        with_spans=with_spans,
        piece_encoder=piece_encoder,
        transformer=transformer,
    )


def build_transformer_model_v1(
    *,
    with_spans: Callable[
        [TorchTransformerModelT],
        SpanExtractorModelT,
    ],
    transformer: TorchTransformerModelT,
    piece_encoder: Tok2PiecesModelT,
) -> TransformerModelT:
    # FIXME: do we want to make `remove_bos_eos` configurable as well or
    #        is it always the same post-processing?
    layers = [
        with_non_ws_tokens(
            chain(piece_encoder, with_spans(transformer), remove_bos_eos())
        )
    ]
    refs = {
        "piece_encoder": piece_encoder,
        "transformer": transformer,
    }
    return Model(
        "transformer_model",
        transformer_model_forward,
        init=transformer_model_init,
        layers=layers,
        refs=refs,  # type: ignore
        attrs={
            "replace_listener": _replace_listener,
            "replace_listener_cfg": _replace_listener_cfg,
        },
    )


def transformer_model_forward(
    model: TransformerModelT, docs: TransformerInT, is_train: bool
) -> Tuple[TransformerOutT, TransformerBackpropT]:
    Y, backprop_layer = model.layers[0](docs, is_train=is_train)

    def backprop(dY):
        backprop_layer(dY)

        # Return empty list for backprop, since we cannot backprop into piece
        # identifiers.
        return []

    return Y, backprop


def transformer_model_init(
    model: TransformerModelT, X: Optional[TransformerInT] = None, Y=None
) -> Model:
    model.layers[0].initialize(X, Y)
    return model


def _pytorch_encoder(
    encoder: CuratedEncoderT,
    *,
    mixed_precision: bool = False,
    grad_scaler_config: dict = SimpleFrozenDict(),
) -> TorchTransformerModelT:
    if isinstance(grad_scaler_config, SimpleFrozenDict):
        # Create a new, mutable dict instance.
        grad_scaler_config = {}

    if "enabled" not in grad_scaler_config:
        grad_scaler_config["enabled"] = mixed_precision

    model = PyTorchWrapper_v2(
        CuratedTransformer(encoder),
        convert_inputs=partial(
            _convert_inputs,
            max_model_seq_len=encoder.max_seq_len,
            padding_idx=encoder.padding_idx,
        ),
        convert_outputs=_convert_outputs,
        mixed_precision=mixed_precision,
        grad_scaler=PyTorchGradScaler(**grad_scaler_config),
    )

    # This attribute is set by the parent Pipe instance before each forward pass.
    model.attrs["_all_layer_outputs"] = True

    return model


def _torchscript_encoder(*, model_max_length: int, padding_idx: int) -> Model:
    return TorchScriptWrapper_v1(
        convert_inputs=partial(
            _convert_inputs,
            max_model_seq_len=model_max_length,
            padding_idx=padding_idx,
        ),
        convert_outputs=_convert_outputs,
    )


def _convert_inputs(
    model: Model,
    X: TorchTransformerInT,
    is_train: bool,
    *,
    max_model_seq_len: int,
    padding_idx: int,
) -> Tuple[ArgsKwargs, Callable[[Any], List[Ints1d]]]:
    ops = model.ops
    max_seq_len = max(x.size for x in X)
    if max_seq_len > max_model_seq_len:
        raise ValueError(
            Errors.E009.format(seq_len=max_seq_len, max_seq_len=max_model_seq_len)
        )

    # Transform the list of strided spans to a padded array.
    Xt = ops.xp.full((len(X), max_seq_len), padding_idx)
    for i in range(len(X)):
        span = X[i]
        span_len = span.shape[0]
        Xt[i, :span_len] = span
    Xt = xp2torch(Xt)

    def convert_from_torch_backward(d_inputs: Any):
        # No gradients for the inputs.
        return [ops.alloc1f(x.shape[0]) for x in X]

    output = ArgsKwargs(args=(Xt,), kwargs={})
    return output, convert_from_torch_backward


def _convert_outputs(
    model: Model,
    inputs_outputs: Tuple[TorchTransformerInT, PyTorchTransformerOutput],
    is_train: bool,
) -> Tuple[TorchTransformerOutT, Callable[[List[List[Floats2d]]], ArgsKwargs]]:
    model_inputs, model_outputs = inputs_outputs
    ops = model.ops
    all_layer_outputs: bool = model.attrs["_all_layer_outputs"]

    # Strip the outputs for the padding timesteps
    # while preserving all the outputs (and their order).
    input_lens = [x.shape[0] for x in model_inputs]
    if all_layer_outputs:
        Yt = [
            [output[i, :len, :] for output in model_outputs.all_outputs]
            for i, len in enumerate(input_lens)
        ]
    else:
        Yt = [
            [model_outputs.all_outputs[-1][i, :len, :]]
            for i, len in enumerate(input_lens)
        ]

    Y = [
        [cast(Floats2d, torch2xp(layer, ops=ops)) for layer in output] for output in Yt
    ]
    output = TransformerModelOutput(outputs=Y, last_layer_only=not all_layer_outputs)

    def convert_for_torch_backward(dY: List[List[Floats2d]]):
        Yt_flat = [y for inner in Yt for y in inner]
        dYt_flat = [xp2torch(y) for inner in dY for y in inner]
        assert len(Yt_flat) == len(dYt_flat)

        return ArgsKwargs(
            args=(Yt_flat,),
            kwargs={"grad_tensors": dYt_flat},
        )

    return output, convert_for_torch_backward


def _replace_listener(trf_model):
    raise ValueError(Errors.E010)


def _replace_listener_cfg(trf_model_cfg, listener_model_cfg):
    raise ValueError(Errors.E010)


def build_pytorch_checkpoint_loader_v1(
    *, path: Path
) -> Callable[
    [TorchTransformerModelT, Optional[List[Doc]], Optional[List[Doc]]],
    TorchTransformerModelT,
]:
    """Construct a callback that initializes a supported transformer
    model with weights from a PyTorch checkpoint.

    path (Path):
        Path to the PyTorch checkpoint.
    """

    def load(model, X=None, Y=None):
        encoder = model.shims[0]._model
        device = get_torch_default_device()
        params = torch.load(path, map_location=device)
        params = convert_pretrained_model_for_encoder(encoder, params)
        encoder.load_state_dict(params)
        encoder.to(device)
        return model

    return load



from typing import Dict, Any
from functools import partial
import pytest
import spacy
from spacy import Config, util
from spacy.training import Example
from spacy.training.initialize import init_nlp
from spacy.training.loop import train
from spacy.language import Language
from spacy.util import registry as spacy_registry
import torch

from curated_transformers.models.architectures import (
    build_camembert_transformer_model_v1,
    build_xlmr_transformer_model_v1,
    build_bert_transformer_model_v1,
    build_roberta_transformer_model_v1,
)
from curated_transformers.models.hf_loader import build_hf_transformer_encoder_loader_v1
from curated_transformers.models.with_strided_spans import (
    build_with_strided_spans_v1,
)
from curated_transformers.tokenization import (
    build_bert_wordpiece_encoder_v1,
    build_byte_bpe_encoder_v1,
    build_camembert_sentencepiece_encoder_v1,
    build_hf_piece_encoder_loader_v1,
    build_xlmr_sentencepiece_encoder_v1,
)
from curated_transformers.pipeline.transformer import make_transformer
from curated_transformers.util import create_gradual_transformer_unfreezing
from curated_transformers._compat import has_hf_transformers, transformers

from ..util import torch_assertclose


cfg_string_last_layer_listener = """
    # LastTransformerLayerListener

    [nlp]
    lang = "en"
    pipeline = ["transformer","tagger"]

    [components]

    [components.tagger]
    factory = "tagger"

    [components.tagger.model]
    @architectures = "spacy.Tagger.v2"
    nO = null

    [components.tagger.model.tok2vec]
    @architectures = "curated-transformers.LastTransformerLayerListener.v1"
    width = ${components.transformer.model.hidden_width}
    pooling = {"@layers":"reduce_mean.v1"}

    [components.transformer]
    factory = "curated_transformer"
    all_layer_outputs = False

    [components.transformer.model]
    @architectures = "curated-transformers.BertTransformer.v1"
    vocab_size = 28996
    num_hidden_layers = 1
    hidden_width = 60
    piece_encoder = {"@architectures":"curated-transformers.BertWordpieceEncoder.v1"}
    with_spans = {"@architectures":"curated-transformers.WithStridedSpans.v1"}

    [initialize]

    [initialize.components]

    [initialize.components.transformer]

    [initialize.components.transformer.piecer_loader]
    @model_loaders = "curated-transformers.HFPieceEncoderLoader.v1"
    name = "bert-base-cased"
"""

cfg_string_scalar_weighting_layer_listener = """
    # ScalarWeightingListener

    [nlp]
    lang = "en"
    pipeline = ["transformer","tagger"]

    [components]

    [components.tagger]
    factory = "tagger"

    [components.tagger.model]
    @architectures = "spacy.Tagger.v2"
    nO = null

    [components.tagger.model.tok2vec]
    @architectures = "curated-transformers.ScalarWeightingListener.v1"
    width = ${components.transformer.model.hidden_width}
    pooling = {"@layers":"reduce_mean.v1"}

    [components.tagger.model.tok2vec.weighting]
    @architectures = "curated-transformers.ScalarWeight.v1"
    num_layers = ${components.transformer.model.num_hidden_layers}

    [components.transformer]
    factory = "curated_transformer"
    all_layer_outputs = True

    [components.transformer.model]
    @architectures = "curated-transformers.BertTransformer.v1"
    vocab_size = 28996
    num_hidden_layers = 1
    hidden_width = 60
    piece_encoder = {"@architectures":"curated-transformers.BertWordpieceEncoder.v1"}
    with_spans = {"@architectures":"curated-transformers.WithStridedSpans.v1"}

    [initialize]

    [initialize.components]

    [initialize.components.transformer]

    [initialize.components.transformer.piecer_loader]
    @model_loaders = "curated-transformers.HFPieceEncoderLoader.v1"
    name = "bert-base-cased"
"""

TRAIN_DATA = [
    (
        "I like green eggs",
        {"tags": ["N", "V", "J", "N"], "cats": {"preference": 1.0, "imperative": 0.0}},
    ),
    (
        "Eat blue ham",
        {"tags": ["V", "J", "N"], "cats": {"preference": 0.0, "imperative": 1.0}},
    ),
]


def create_and_train_tagger(cfg_string):
    config = Config().from_str(cfg_string)
    nlp = util.load_model_from_config(config, auto_fill=True, validate=True)
    tagger = nlp.get_pipe("tagger")

    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))
        for tag in t[1]["tags"]:
            tagger.add_label(tag)

    optimizer = nlp.initialize(lambda: train_examples)

    for _ in range(10):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)

    return nlp


def evaluate_tagger_on_train_data(model):
    docs = list(model.pipe(["Eat blue ham", "I like green eggs"]))
    assert [t.tag_ for t in docs[0]] == ["V", "J", "N"]
    assert [t.tag_ for t in docs[1]] == ["N", "V", "J", "N"]


def test_default_pipe_config_can_be_constructed():
    nlp = spacy.blank("en")
    nlp.add_pipe("curated_transformer", config={"model": {"vocab_size": 32}})


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize(
    "cfg_string",
    [cfg_string_last_layer_listener, cfg_string_scalar_weighting_layer_listener],
)
def test_tagger(cfg_string):
    model = create_and_train_tagger(cfg_string)
    evaluate_tagger_on_train_data(model)


def _hf_tokenize_per_token(tokenizer, docs, *, roberta=False):
    if roberta:
        hf_encoding = [
            tokenizer(
                [
                    doc[idx - 1].whitespace_ + token.text if idx > 0 else token.text
                    for idx, token in enumerate(doc)
                ]
            )
            for doc in docs
        ]
    else:
        hf_encoding = [tokenizer([token.text for token in doc]) for doc in docs]
    ids = []
    lens = []
    bos_id = (
        tokenizer.bos_token_id
        if tokenizer.bos_token_id is not None
        else tokenizer.cls_token_id
    )
    eos_id = (
        tokenizer.eos_token_id
        if tokenizer.eos_token_id is not None
        else tokenizer.sep_token_id
    )
    for i in range(len(hf_encoding)):
        doc_ids = [id for e in hf_encoding[i]["input_ids"] for id in e[1:-1]]
        ids.append([bos_id] + doc_ids + [eos_id])
        lens.append(len(ids[-1]))

    torch_ids = torch.full(
        (len(ids), max(lens)), tokenizer.pad_token_id, dtype=torch.int
    )
    for i in range(len(ids)):
        torch_ids[i][: len(ids[i])] = torch.tensor(ids[i])

    attention_mask = torch_ids.ne(tokenizer.pad_token_id)

    return torch_ids, attention_mask, lens


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_bert_transformer_pipe_against_hf():
    nlp = spacy.blank("en")
    model = build_bert_transformer_model_v1(
        piece_encoder=build_bert_wordpiece_encoder_v1(),
        with_spans=build_with_strided_spans_v1(),
        vocab_size=28996,
    )
    model.get_ref("transformer").init = build_hf_transformer_encoder_loader_v1(
        name="bert-base-cased"
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name="bert-base-cased"
    )
    model.initialize()
    pipe = make_transformer(nlp, "transformer", model)

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("bert-base-cased")
    hf_model = transformers.AutoModel.from_pretrained("bert-base-cased")

    docs = [
        nlp.make_doc("I saw a girl with a telescope."),
        nlp.make_doc("Today we will eat poké bowl."),
    ]

    hf_ids, attention_mask, lens = _hf_tokenize_per_token(hf_tokenizer, docs)
    hf_encoding = hf_model(hf_ids, attention_mask=attention_mask)
    docs = list(pipe.pipe(docs))

    for doc, hf_doc_encoding, encoding_len in zip(
        docs, hf_encoding.last_hidden_state, lens
    ):
        torch_assertclose(
            hf_doc_encoding[:encoding_len][1:-1],
            torch.tensor(doc._.trf_data.last_hidden_layer_state.dataXd),
        )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_camembert_transformer_pipe_against_hf():
    nlp = spacy.blank("fr")
    model = build_camembert_transformer_model_v1(
        piece_encoder=build_camembert_sentencepiece_encoder_v1(),
        with_spans=build_with_strided_spans_v1(),
        vocab_size=32005,
    )
    model.get_ref("transformer").init = build_hf_transformer_encoder_loader_v1(
        name="camembert-base"
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name="camembert-base"
    )
    model.initialize()
    pipe = make_transformer(nlp, "transformer", model)

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("camembert-base")
    hf_model = transformers.AutoModel.from_pretrained("camembert-base")

    docs = [
        nlp.make_doc("J'ai vu une fille avec un télescope."),
        nlp.make_doc("Aujourd'hui, nous allons manger un poké bowl."),
    ]

    hf_ids, attention_mask, lens = _hf_tokenize_per_token(hf_tokenizer, docs)
    hf_encoding = hf_model(hf_ids, attention_mask=attention_mask)
    docs = list(pipe.pipe(docs))

    for doc, hf_doc_encoding, encoding_len in zip(
        docs, hf_encoding.last_hidden_state, lens
    ):
        torch_assertclose(
            hf_doc_encoding[:encoding_len][1:-1],
            torch.tensor(doc._.trf_data.last_hidden_layer_state.dataXd),
        )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_roberta_transformer_pipe_against_hf():
    nlp = spacy.blank("en")
    model = build_roberta_transformer_model_v1(
        piece_encoder=build_byte_bpe_encoder_v1(),
        with_spans=build_with_strided_spans_v1(),
        vocab_size=50265,
    )
    model.get_ref("transformer").init = build_hf_transformer_encoder_loader_v1(
        name="roberta-base"
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name="roberta-base"
    )
    model.initialize()
    pipe = make_transformer(nlp, "transformer", model)

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("roberta-base")
    hf_model = transformers.AutoModel.from_pretrained("roberta-base")

    docs = [
        nlp.make_doc("I saw a girl with a telescope."),
        nlp.make_doc("Today we will eat poké bowl."),
    ]

    hf_ids, attention_mask, lens = _hf_tokenize_per_token(
        hf_tokenizer, docs, roberta=True
    )
    hf_encoding = hf_model(hf_ids, attention_mask=attention_mask)
    docs = list(pipe.pipe(docs))

    for doc, hf_doc_encoding, encoding_len in zip(
        docs, hf_encoding.last_hidden_state, lens
    ):
        torch_assertclose(
            hf_doc_encoding[:encoding_len][1:-1],
            torch.tensor(doc._.trf_data.last_hidden_layer_state.dataXd),
        )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_xlmr_transformer_pipe_against_hf():
    nlp = spacy.blank("en")
    model = build_xlmr_transformer_model_v1(
        piece_encoder=build_xlmr_sentencepiece_encoder_v1(),
        with_spans=build_with_strided_spans_v1(),
        vocab_size=250002,
    )
    model.get_ref("transformer").init = build_hf_transformer_encoder_loader_v1(
        name="xlm-roberta-base"
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name="xlm-roberta-base"
    )
    model.initialize()
    pipe = make_transformer(nlp, "transformer", model)

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("xlm-roberta-base")
    hf_model = transformers.AutoModel.from_pretrained("xlm-roberta-base")

    docs = [
        nlp.make_doc("I saw a girl with a telescope."),
        nlp.make_doc("Today we will eat poké bowl."),
    ]

    hf_ids, attention_mask, lens = _hf_tokenize_per_token(hf_tokenizer, docs)
    hf_encoding = hf_model(hf_ids, attention_mask=attention_mask)
    docs = list(pipe.pipe(docs))

    for doc, hf_doc_encoding, encoding_len in zip(
        docs, hf_encoding.last_hidden_state, lens
    ):
        torch_assertclose(
            hf_doc_encoding[:encoding_len][1:-1],
            torch.tensor(doc._.trf_data.last_hidden_layer_state.dataXd),
        )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_frozen_transformer_pipe():
    config = Config().from_str(cfg_string_scalar_weighting_layer_listener)
    nlp = util.load_model_from_config(config, auto_fill=True, validate=True)
    tagger = nlp.get_pipe("tagger")
    transformer = nlp.get_pipe("transformer")
    transformer.frozen = True

    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))
        for tag in t[1]["tags"]:
            tagger.add_label(tag)

    optimizer = nlp.initialize(lambda: train_examples)

    def get_transformer_params_sorted():
        params = transformer.model.get_ref("transformer").shims[0]._model.state_dict()
        return list(sorted(params.items()))

    transformer_init_params = [
        (k, v.clone()) for k, v in get_transformer_params_sorted()
    ]

    for i in range(5):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)

    transformer_trained_params = get_transformer_params_sorted()
    for (old_param, old_vec), (new_param, new_vec) in zip(
        transformer_init_params, transformer_trained_params
    ):
        assert old_param == new_param
        torch_assertclose(
            old_vec,
            new_vec,
        )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_transformer_pipe_outputs():
    nlp = spacy.blank("en")
    model = build_xlmr_transformer_model_v1(
        piece_encoder=build_xlmr_sentencepiece_encoder_v1(),
        with_spans=build_with_strided_spans_v1(),
        vocab_size=250002,
    )
    model.get_ref("transformer").init = build_hf_transformer_encoder_loader_v1(
        name="xlm-roberta-base"
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name="xlm-roberta-base"
    )
    model.initialize()
    pipe = make_transformer(nlp, "transformer", model, all_layer_outputs=False)

    docs = [
        nlp.make_doc("I saw a girl with a telescope."),
        nlp.make_doc("Today we will eat poké bowl."),
    ]
    docs = list(pipe.pipe(docs))
    assert all([doc._.trf_data.last_layer_only for doc in docs]) == True
    assert all([len(doc._.trf_data.all_outputs) == 1 for doc in docs]) == True

    pipe = make_transformer(nlp, "transformer", model, all_layer_outputs=True)
    docs = list(pipe.pipe(docs))
    assert all([not doc._.trf_data.last_layer_only for doc in docs]) == True
    assert all([len(doc._.trf_data.all_outputs) == 12 + 1 for doc in docs]) == True


cfg_string_gradual_unfreezing = (
    cfg_string_last_layer_listener
    + """
    [corpora]

    [corpora.train]
    @readers = "spacy.Corpus.v1"
    path = "toy-en-corpus.spacy"
    max_length = 500
    gold_preproc = false
    limit = 0

    [corpora.dev]
    @readers = "spacy.Corpus.v1"
    path = "toy-en-corpus.spacy"
    max_length = 500
    gold_preproc = false
    limit = 0

    [training]
    train_corpus = "corpora.train"
    dev_corpus = "corpora.dev"
    seed = 1
    gpu_allocator = "pytorch"
    dropout = 0.1
    accumulate_gradient = 3
    patience = 5000
    max_epochs = 1
    max_steps = 6
    eval_frequency = 10

    [training.batcher]
    @batchers = "spacy.batch_by_padded.v1"
    discard_oversize = False
    get_length = null
    size = 1
    buffer = 256
"""
)


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_gradual_transformer_unfreezing(test_dir):
    def wrapped_callback():
        def inner(
            nlp: Language,
            args: Dict[str, Any],
            gradual_unfreezing_callback,
            unfreeze_step,
        ):
            current_step = args["step"]
            gradual_unfreezing_callback(nlp, args)

            transformer = nlp.get_pipe("transformer")
            if current_step < unfreeze_step:
                assert transformer.frozen == True
            else:
                assert transformer.frozen == False

        return partial(
            inner,
            gradual_unfreezing_callback=create_gradual_transformer_unfreezing({"*": 3}),
            unfreeze_step=3,
        )

    spacy_registry.callbacks.register(
        "test_gradual_unfreezing_callback",
        func=wrapped_callback,
    )

    config = Config().from_str(cfg_string_gradual_unfreezing, interpolate=False)
    config["corpora"]["train"]["path"] = str(test_dir / "toy-en-corpus.spacy")
    config["corpora"]["dev"]["path"] = str(test_dir / "toy-en-corpus.spacy")
    config["training"]["before_update"] = {
        "@callbacks": "test_gradual_unfreezing_callback"
    }
    nlp = util.load_model_from_config(config, auto_fill=True, validate=True)

    nlp = init_nlp(config)
    train(nlp)
    assert nlp.get_pipe("transformer").frozen == False

    with pytest.raises(ValueError):
        create_gradual_transformer_unfreezing({"transformer": 5, "*": 4})



from typing import Optional
from torch.nn import Module
from torch import Tensor

from ..bert import BertEmbeddings, BertEmbeddingConfig, BertLayerConfig


class RobertaEmbeddings(Module):
    def __init__(
        self,
        embedding_config: BertEmbeddingConfig,
        layer_config: BertLayerConfig,
        *,
        padding_idx: int
    ) -> None:
        super().__init__()

        self.inner = BertEmbeddings(embedding_config, layer_config)
        self.padding_idx = padding_idx

    def _get_position_ids(self, x: Tensor) -> Tensor:
        # We need to generate the position IDs from the
        # input tensor to pass to the embedding layer and
        # handle padding, c.f https://github.com/huggingface/transformers/blob/330247ede2d8265aae9ab0b7a0d1a811c344960d/src/transformers/models/roberta/modeling_roberta.py#L1566

        mask = x.ne(self.padding_idx).int()
        return (mask.cumsum(dim=1) * mask) + self.padding_idx

    def forward(
        self,
        input_ids: Tensor,
        token_type_ids: Optional[Tensor] = None,
        position_ids: Optional[Tensor] = None,
    ) -> Tensor:
        """
        Shapes:
            input_ids, token_type_ids, position_ids - (batch, seq_len)
        """
        if position_ids is None:
            position_ids = self._get_position_ids(input_ids)

        return self.inner(input_ids, token_type_ids, position_ids)



from typing import Optional
import torch
from torch import Tensor
from torch.nn import Dropout, Embedding, LayerNorm, Linear, Module

from .config import BertEmbeddingConfig, BertLayerConfig


class BertEmbeddings(Module):
    def __init__(
        self, embedding_config: BertEmbeddingConfig, layer_config: BertLayerConfig
    ) -> None:
        super().__init__()

        self.word_embeddings = Embedding(
            num_embeddings=embedding_config.vocab_size,
            embedding_dim=embedding_config.embedding_width,
        )
        self.token_type_embeddings = Embedding(
            num_embeddings=embedding_config.type_vocab_size,
            embedding_dim=embedding_config.embedding_width,
        )
        self.position_embeddings = Embedding(
            num_embeddings=embedding_config.max_position_embeddings,
            embedding_dim=embedding_config.embedding_width,
        )

        if embedding_config.embedding_width != layer_config.hidden_width:
            self.projection = Linear(
                embedding_config.embedding_width, layer_config.hidden_width
            )
        else:
            self.projection = None  # type: ignore

        self.layer_norm = LayerNorm(
            embedding_config.embedding_width, eps=embedding_config.layer_norm_eps
        )
        self.dropout = Dropout(p=embedding_config.dropout_prob)

    def _get_position_ids(self, x: Tensor) -> Tensor:
        return torch.arange(x.shape[1], device=x.device).expand(1, -1)

    def _get_token_type_ids(self, x: Tensor) -> Tensor:
        return torch.zeros_like(x)

    def forward(
        self,
        input_ids: Tensor,
        token_type_ids: Optional[Tensor] = None,
        position_ids: Optional[Tensor] = None,
    ) -> Tensor:
        """
        Shapes:
            input_ids, token_type_ids, position_ids - (batch, seq_len)
        """
        if token_type_ids is None:
            token_type_ids = self._get_token_type_ids(input_ids)
        if position_ids is None:
            position_ids = self._get_position_ids(input_ids)

        input_embeddings = self.word_embeddings(input_ids)
        position_embeddings = self.position_embeddings(position_ids)
        token_type_embeddings = self.token_type_embeddings(token_type_ids)

        embedding_sum = input_embeddings
        embedding_sum += token_type_embeddings
        embedding_sum += position_embeddings
        normalized = self.layer_norm(embedding_sum)
        embeddings = self.dropout(normalized)

        if self.projection is not None:
            return self.projection(embeddings)
        else:
            return embeddings



from typing import Optional

import torch
from torch.nn import Module
from torch import Tensor

from ..attention import AttentionMask
from ..bert.layer import BertEncoderLayer
from ...output import PyTorchTransformerOutput
from .embeddings import RobertaEmbeddings
from .config import RobertaConfig


class RobertaEncoder(Module):
    def __init__(self, config: RobertaConfig):
        super().__init__()

        self.embeddings = RobertaEmbeddings(
            config.embedding, config.layer, padding_idx=config.padding_idx
        )
        self.padding_idx = config.padding_idx
        self.max_seq_len = config.model_max_length
        self.layers = torch.nn.ModuleList(
            [
                BertEncoderLayer(config.layer, config.attention)
                for _ in range(config.layer.num_hidden_layers)
            ]
        )

    def _create_attention_mask(self, x: Tensor) -> AttentionMask:
        return AttentionMask(x.ne(self.padding_idx))

    def forward(
        self,
        input_ids: Tensor,
        attention_mask: Optional[AttentionMask] = None,
        token_type_ids: Optional[Tensor] = None,
    ) -> PyTorchTransformerOutput:
        """
        Shapes:
            input_ids, attention_mask, token_type_ids - (batch, seq_len)
        """
        if attention_mask is None:
            attention_mask = self._create_attention_mask(input_ids)

        embeddings = self.embeddings(input_ids, token_type_ids, None)
        layer_output = embeddings

        layer_outputs = []
        for layer in self.layers:
            layer_output = layer(layer_output, attention_mask)
            layer_outputs.append(layer_output)

        return PyTorchTransformerOutput(
            embedding_output=embeddings, layer_hidden_states=layer_outputs
        )



from typing import Callable, Optional, Tuple
from pathlib import Path

from cutlery import ByteBPEProcessor
import srsly
from thinc.api import Model, Ragged, deserialize_attr, serialize_attr

from .types import (
    Tok2PiecesBackpropT,
    Tok2PiecesInT,
    Tok2PiecesModelT,
    Tok2PiecesOutT,
)
from ..errors import Errors


@serialize_attr.register(ByteBPEProcessor)
def serialize_byte_bpe_processor(_, value: ByteBPEProcessor, name: str, model) -> bytes:
    data = {"merges": value.merges, "vocab": value.vocab}
    return srsly.msgpack_dumps(data)


@deserialize_attr.register(ByteBPEProcessor)
def deserialize_byte_bpe_processor(
    _, value: bytes, name: str, model
) -> ByteBPEProcessor:
    data = srsly.msgpack_loads(value)
    return ByteBPEProcessor(data["vocab"], data["merges"])


def build_byte_bpe_encoder_v1() -> Tok2PiecesModelT:
    """Construct a Byte-BPE piece encoder model that accepts a list
    of token sequences or documents and returns a corresponding list
    of piece identifiers.

    This model must be separately initialized using an appropriate
    loader.
    """
    return Model(
        "byte_bpe_encoder",
        forward=byte_bpe_encoder_forward,
        attrs={
            "byte_bpe_processor": ByteBPEProcessor({}, []),
            "unk_piece": "<unk>",
            "bos_piece": "<s>",
            "eos_piece": "</s>",
        },
    )


def byte_bpe_encoder_forward(
    model: Model, X: Tok2PiecesInT, is_train: bool
) -> Tuple[Tok2PiecesOutT, Tok2PiecesBackpropT]:
    bbp: ByteBPEProcessor = model.attrs["byte_bpe_processor"]
    bos_piece: str = model.attrs["bos_piece"]
    eos_piece: str = model.attrs["eos_piece"]
    unk_piece: str = model.attrs["unk_piece"]
    bos_id = bbp.piece_id(bos_piece)
    if bos_id is None:
        raise ValueError(Errors.E019.format(piece="BOS"))
    eos_id = bbp.piece_id(eos_piece)
    if eos_id is None:
        raise ValueError(Errors.E019.format(piece="EOS"))
    unk_id = bbp.piece_id(unk_piece)
    if unk_id is None:
        raise ValueError(Errors.E019.format(piece="UNK"))

    pieces = []
    for doc in X:
        # TODO: check whether as single bos/eos per doc is what we want.
        #   The issue is that we probably do not have sentence
        #   boundaries yet, since they are predicted by a pipe.
        doc_pieces = [bos_id]
        lens = [1]

        for idx, token in enumerate(doc):
            # GPT-2/RoBERTa tokenization preserves preceding space character.
            if idx > 0:
                text = doc[idx - 1].whitespace_ + token.text
            else:
                text = token.text

            piece_ids = bbp.encode_as_ids(text)

            doc_pieces.extend(piece_ids)
            lens.append(len(piece_ids))

        doc_pieces.append(eos_id)
        lens.append(1)
        pieces.append(
            Ragged(
                model.ops.asarray1i(doc_pieces),
                model.ops.asarray1i(lens),
            )
        )

    return pieces, lambda dY: []


def build_byte_bpe_encoder_loader_v1(
    *, vocab_path: Path, merges_path: Path
) -> Callable[
    [Tok2PiecesModelT, Optional[Tok2PiecesInT], Optional[Tok2PiecesInT]],
    Tok2PiecesModelT,
]:
    """Construct a callback that initializes a Byte-BPE piece encoder
    model.

    vocab_path (Path):
        Path to the vocabulary file.
    merges_path (Path):
        Path to the merges file.
    """

    def load(model, X=None, Y=None):
        model.attrs["byte_bpe_processor"] = ByteBPEProcessor.load_from_files(
            vocab=vocab_path, merges=merges_path
        )
        return model

    return load



from typing import Callable, Dict, Optional, OrderedDict, Tuple
from pathlib import Path
from thinc.api import Model, Ragged
import unicodedata

from .types import Tok2PiecesBackpropT, Tok2PiecesInT, Tok2PiecesModelT, Tok2PiecesOutT
from ..errors import Errors


def build_char_encoder_v1() -> Tok2PiecesModelT:
    """Construct a character piece encoder model that accepts a list
    of token sequences or documents and returns a corresponding list
    of piece identifiers.

    This model must be separately initialized using an appropriate
    loader.

    normalize (str):
       Unicode normalization to apply before encoding a token. Defaults to
       "NFKC".
    """
    return Model(
        "char_encoder",
        forward=char_encoder_forward,
        attrs={
            "bos_piece": "[CLS]",
            "eos_piece": "[SEP]",
            "unk_piece": "[UNK]",
            "normalize": "NFKC",
            "vocab": None,
        },
    )


def char_encoder_forward(
    model: Model, X: Tok2PiecesInT, is_train: bool
) -> Tuple[Tok2PiecesOutT, Tok2PiecesBackpropT]:
    """Construct a character piece encoder model that accepts a list
    of token sequences or documents and returns a corresponding list
    of piece identifiers.

    This model must be separately initialized using an appropriate
    loader.
    """
    vocab: Optional[Dict[str, int]] = model.attrs["vocab"]
    if vocab is None:
        raise ValueError(Errors.E020)

    bos_piece: str = model.attrs["bos_piece"]
    eos_piece: str = model.attrs["eos_piece"]
    unk_piece: str = model.attrs["unk_piece"]
    normalize: Optional[str] = model.attrs["normalize"]
    bos_id = vocab[bos_piece]
    eos_id = vocab[eos_piece]
    unk_id = vocab[unk_piece]

    pieces = []
    for doc in X:
        doc_pieces = [bos_id]
        lens = [1]

        for token in doc:
            text = (
                unicodedata.normalize(normalize, token.text)
                if normalize is not None
                else token.text
            )
            # Most encoders will mark a full token as unknown. The character
            # encoder behaves differently, only replacing unknown characters
            # by the unk id.
            piece_ids = [vocab.get(char, unk_id) for char in text]
            doc_pieces.extend(piece_ids)
            lens.append(len(piece_ids))

        doc_pieces.append(eos_id)
        lens.append(1)
        pieces.append(
            Ragged(model.ops.asarray1i(doc_pieces), model.ops.asarray1i(lens))
        )

    return pieces, lambda dY: []


def build_char_encoder_loader_v1(
    *,
    path: Path,
    bos_piece: str = "[CLS]",
    eos_piece: str = "[SEP]",
    unk_piece: str = "[UNK]",
    normalize: Optional[str] = "NFKC",
) -> Callable[
    [Tok2PiecesModelT, Optional[Tok2PiecesInT], Optional[Tok2PiecesInT]],
    Tok2PiecesModelT,
]:
    """Construct a callback that initializes a character piece encoder
    model.

    path (Path):
        Path to the serialized character model.
    """

    def load(model, X=None, Y=None):
        if model.name != "char_encoder":
            raise ValueError(Errors.E021.format(model_name=model.name))

        model.attrs["bos_piece"] = bos_piece
        model.attrs["eos_piece"] = eos_piece
        model.attrs["unk_piece"] = unk_piece
        model.attrs["normalize"] = normalize

        vocab = OrderedDict()
        with open(path, encoding="utf-8") as f:
            for char in f:
                char = char.rstrip("\r\n")
                if normalize is not None:
                    char = unicodedata.normalize(normalize, char)
                vocab[char] = len(vocab)
        model.attrs["vocab"] = vocab
        return model

    return load



import math
import torch
from torch import Tensor
from torch.nn import Module


class GeluNew(Module):
    """GELU approximation, called `gelu_new` in many transformer models."""

    def forward(self, input: Tensor) -> Tensor:
        """
        Shapes:
            input - (batch, seq_len, width)
        """
        return (
            0.5
            * input
            * (
                1.0
                + torch.tanh(
                    math.sqrt(2.0 / math.pi)
                    * (input + 0.044715 * torch.pow(input, 3.0))
                )
            )
        )



from thinc.api import Ragged, reduce_sum, NumpyOps
from thinc.util import convert_recursive, is_cupy_array

from curated_transformers.models.pooling import (
    with_ragged_last_layer,
    with_ragged_layers,
)


def test_with_ragged_last_layer():
    pooler = with_ragged_last_layer(reduce_sum())
    ops = pooler.ops

    docs = [
        Ragged(ops.asarray1f([1.0, 2.0, 3.0]), lengths=ops.asarray1i([1, 2])),
        Ragged(ops.asarray1f([4.0, 5.0, 6.0]), lengths=ops.asarray1i([2, 1])),
        Ragged(ops.asarray1f([]), lengths=ops.asarray1i([])),
    ]

    Y, backprop = pooler(docs, is_train=True)

    for y, y_h in zip(
        Y,
        [
            ops.asarray2f([[1.0], [5.0]]),
            ops.asarray2f([[9.0], [6.0]]),
            ops.asarray2f([]).reshape(0, 1),
        ],
    ):
        ops.xp.testing.assert_array_equal(y, y_h)

    dX = backprop(
        [
            ops.asarray2f([[1.0], [-1.0]]),
            ops.asarray2f([[-1.0], [1.0]]),
            ops.asarray2f([]).reshape(0, 1),
        ],
    )

    assert len(dX) == 3

    ops.xp.testing.assert_array_equal(
        dX[0].dataXd, ops.asarray2f([[1.0], [-1.0], [-1.0]])
    )
    ops.xp.testing.assert_array_equal(dX[0].lengths, ops.asarray1i([1, 2]))

    ops.xp.testing.assert_array_equal(
        dX[1].dataXd, ops.asarray2f([[-1.0], [-1.0], [1.0]])
    )
    ops.xp.testing.assert_array_equal(dX[1].lengths, ops.asarray1i([2, 1]))

    ops.xp.testing.assert_array_equal(dX[2].dataXd, ops.asarray2f([]).reshape(0, 1))
    ops.xp.testing.assert_array_equal(dX[2].lengths, ops.asarray1i([]))


def test_with_ragged_layers():
    pooler = with_ragged_layers(reduce_sum())
    ops = pooler.ops
    numpy = NumpyOps().xp

    docs = [
        [
            Ragged(ops.asarray1f([1.0, 2.0, 3.0]), lengths=ops.asarray1i([1, 2])),
            Ragged(ops.asarray1f([3.0, 2.0, 1.0]), lengths=ops.asarray1i([1, 2])),
        ],
        [
            Ragged(ops.asarray1f([4.0, 5.0, 6.0]), lengths=ops.asarray1i([2, 1])),
            Ragged(ops.asarray1f([6.0, 5.0, 4.0]), lengths=ops.asarray1i([2, 1])),
        ],
        [
            Ragged(ops.asarray1f([]), lengths=ops.asarray1i([])),
            Ragged(ops.asarray1f([]), lengths=ops.asarray1i([])),
        ],
    ]

    Y, backprop = pooler(docs, is_train=True)
    # Since cupy balks at comparing sequences that mix CPU lists and GPU arrays,
    # we'll need to convert them to numpy (CPU) arrays first.
    Y = convert_recursive(is_cupy_array, lambda a: a.get(), Y)

    numpy.testing.assert_equal(
        Y,
        [
            [
                numpy.array([[1.0], [5.0]]),
                numpy.array([[3.0], [3.0]]),
            ],
            [
                numpy.array([[9.0], [6.0]]),
                numpy.array([[11.0], [4.0]]),
            ],
            [
                numpy.array([]).reshape(0, 1),
                numpy.array([]).reshape(0, 1),
            ],
        ],
    )

    dX = backprop(
        [
            [
                ops.asarray2f([[1.0], [-1.0]]),
                ops.asarray2f([[2.0], [-2.0]]),
            ],
            [
                ops.asarray2f([[-1.0], [1.0]]),
                ops.asarray2f([[-2.0], [2.0]]),
            ],
            [
                ops.asarray2f([]).reshape(0, 1),
                ops.asarray2f([]).reshape(0, 1),
            ],
        ],
    )

    assert len(dX) == 3

    assert len(dX[0]) == 2
    ops.xp.testing.assert_array_equal(
        dX[0][0].dataXd, ops.asarray2f([[1.0], [-1.0], [-1.0]])
    )
    ops.xp.testing.assert_array_equal(dX[0][0].lengths, ops.asarray1i([1, 2]))
    ops.xp.testing.assert_array_equal(
        dX[0][1].dataXd, ops.asarray2f([[2.0], [-2.0], [-2.0]])
    )
    ops.xp.testing.assert_array_equal(dX[0][1].lengths, ops.asarray1i([1, 2]))

    assert len(dX[1]) == 2
    ops.xp.testing.assert_array_equal(
        dX[1][0].dataXd, ops.asarray2f([[-1.0], [-1.0], [1.0]])
    )
    ops.xp.testing.assert_array_equal(
        dX[1][1].dataXd, ops.asarray2f([[-2.0], [-2.0], [2.0]])
    )
    ops.xp.testing.assert_array_equal(dX[1][0].lengths, ops.asarray1i([2, 1]))
    ops.xp.testing.assert_array_equal(dX[1][1].lengths, ops.asarray1i([2, 1]))

    assert len(dX[2]) == 2
    ops.xp.testing.assert_array_equal(dX[2][0].dataXd, ops.asarray2f([]).reshape(0, 1))
    ops.xp.testing.assert_array_equal(dX[2][1].dataXd, ops.asarray2f([]).reshape(0, 1))
    ops.xp.testing.assert_array_equal(dX[2][0].lengths, ops.asarray1i([]))
    ops.xp.testing.assert_array_equal(dX[2][1].lengths, ops.asarray1i([]))



import math
import torch
from torch import Tensor
from torch.nn import Module

# https://pytorch.org/tutorials/beginner/transformer_tutorial.html
class SinusoidalPositionalEmbedding(Module):
    def __init__(self, dim: int, max_len: int, *, normalize=True):
        super().__init__()

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, dim, 2) * (-math.log(10000.0) / dim))

        pe = torch.zeros(max_len, dim)
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)

        if normalize == True:
            l2 = torch.norm(pe, dim=-1)
            pe /= l2.unsqueeze(-1)

        self.pe = pe
        self.pe.requires_grad = False

    def forward(self, x: Tensor) -> Tensor:
        """
        Shapes:
            x - (batch, seq_len)
        """
        return self.pe[x.size(1), :]



from dataclasses import dataclass

from ..bert import BertConfig, BertAttentionConfig, BertEmbeddingConfig
from ..bert import BertLayerConfig


@dataclass
class AlbertLayerConfig(BertLayerConfig):
    inner_group_num: int
    num_hidden_groups: int

    def __init__(self, *args, inner_group_num=1, num_hidden_groups=1, **kwargs):
        super(AlbertLayerConfig, self).__init__(*args, **kwargs)
        self.inner_group_num = inner_group_num
        self.num_hidden_groups = num_hidden_groups


@dataclass
class AlbertConfig(BertConfig):
    layer: AlbertLayerConfig

    def __init__(
        self,
        *,
        embedding_width: int = 128,
        hidden_width: int = 768,
        inner_group_num: int = 1,
        intermediate_width: int = 3072,
        num_attention_heads: int = 12,
        num_hidden_layers: int = 12,
        num_hidden_groups: int = 1,
        attention_probs_dropout_prob: float = 0.0,
        hidden_dropout_prob: float = 0.0,
        hidden_act: str = "gelu_new",
        vocab_size: int = 30000,
        type_vocab_size: int = 2,
        max_position_embeddings: int = 512,
        model_max_length: int = 512,
        layer_norm_eps: float = 1e-12,
        padding_idx: int = 0,
    ):
        self.embedding = BertEmbeddingConfig(
            embedding_width=embedding_width,
            vocab_size=vocab_size,
            type_vocab_size=type_vocab_size,
            max_position_embeddings=max_position_embeddings,
            layer_norm_eps=layer_norm_eps,
            dropout_prob=hidden_dropout_prob,
        )
        self.attention = BertAttentionConfig(
            hidden_width=hidden_width,
            num_attention_heads=num_attention_heads,
            dropout_prob=attention_probs_dropout_prob,
        )
        self.layer = AlbertLayerConfig(
            hidden_width=hidden_width,
            inner_group_num=inner_group_num,
            intermediate_width=intermediate_width,
            num_hidden_layers=num_hidden_layers,
            num_hidden_groups=num_hidden_groups,
            hidden_act=hidden_act,
            layer_norm_eps=layer_norm_eps,
            dropout_prob=hidden_dropout_prob,
        )
        self.model_max_length = model_max_length
        self.padding_idx = padding_idx



from typing import Callable, List, Optional, Tuple
from pathlib import Path
import unicodedata

from cutlery import WordPieceProcessor
from thinc.api import Model, Ragged, deserialize_attr, serialize_attr

from .types import (
    Tok2PiecesBackpropT,
    Tok2PiecesInT,
    Tok2PiecesModelT,
    Tok2PiecesOutT,
)


@serialize_attr.register(WordPieceProcessor)
def serialize_sentencepiece_processor(
    _, value: WordPieceProcessor, name: str, model
) -> bytes:
    return "\n".join(value.to_list()).encode("utf8")


@deserialize_attr.register(WordPieceProcessor)
def deserialize_my_custom_class(
    _, value: bytes, name: str, model
) -> WordPieceProcessor:
    return WordPieceProcessor(value.decode("utf8").split("\n"))


def build_bert_wordpiece_encoder_v1() -> Tok2PiecesModelT:
    """Construct a WordPiece piece encoder model that accepts a list
    of token sequences or documents and returns a corresponding list
    of piece identifiers. This encoder also splits each token
    on punctuation characters, as expected by most BERT models.

    This model must be separately initialized using an appropriate
    loader.
    """
    return Model(
        "wordpiece_encoder",
        forward=wordpiece_encoder_forward,
        attrs={
            "wordpiece_processor": WordPieceProcessor([]),
            "unk_piece": "[UNK]",
            "bos_piece": "[CLS]",
            "eos_piece": "[SEP]",
            "lowercase": False,
            "preprocess": _bert_preprocess,
            "strip_accents": False,
        },
    )


def build_wordpiece_encoder_v1() -> Tok2PiecesModelT:
    """Construct a WordPiece piece encoder model that accepts a list
    of token sequences or documents and returns a corresponding list
    of piece identifiers.

    This model must be separately initialized using an appropriate
    loader.
    """
    return Model(
        "wordpiece_encoder",
        forward=wordpiece_encoder_forward,
        attrs={
            "wordpiece_processor": WordPieceProcessor([]),
            "unk_piece": "[UNK]",
            "bos_piece": "[CLS]",
            "eos_piece": "[SEP]",
            "lowercase": False,
            "preprocess": lambda t: [t],
            "strip_accents": False,
        },
    )


def wordpiece_encoder_forward(
    model: Model, X: Tok2PiecesInT, is_train: bool
) -> Tuple[Tok2PiecesOutT, Tok2PiecesBackpropT]:
    wpp: WordPieceProcessor = model.attrs["wordpiece_processor"]
    bos_piece: str = model.attrs["bos_piece"]
    eos_piece: str = model.attrs["eos_piece"]
    unk_piece: str = model.attrs["unk_piece"]
    lowercase: bool = model.attrs["lowercase"]
    preprocess: Callable[[str], str] = model.attrs["preprocess"]
    strip_accents: bool = model.attrs["strip_accents"]
    bos_id = wpp.get_initial(bos_piece)
    eos_id = wpp.get_initial(eos_piece)
    unk_id = wpp.get_initial(unk_piece)

    pieces = []
    for doc in X:
        # TODO: check whether as single bos/eos per doc is what we want.
        #   The issue is that we probably do not have sentence
        #   boundaries yet, since they are predicted by a pipe.
        doc_pieces = [bos_id]
        lens = [1]

        for token in doc:
            text = token.lower_ if lowercase else token.text
            if strip_accents:
                text = _strip_accents(text)

            piece_ids = [
                unk_id if token_id == -1 else token_id
                for t in preprocess(text)
                for token_id in wpp.encode(t)[0]
            ]

            doc_pieces.extend(piece_ids)
            lens.append(len(piece_ids))

        doc_pieces.append(eos_id)
        lens.append(1)
        pieces.append(
            Ragged(
                model.ops.asarray1i(doc_pieces),
                model.ops.asarray1i(lens),
            )
        )

    return pieces, lambda dY: []


def build_wordpiece_encoder_loader_v1(
    *,
    path: Path,
    bos_piece="[CLS]",
    eos_piece="[SEP]",
    unk_piece="[UNK]",
    lowercase: bool = False,
    strip_accents: bool = False,
) -> Callable[
    [Tok2PiecesModelT, Optional[Tok2PiecesInT], Optional[Tok2PiecesInT]],
    Tok2PiecesModelT,
]:
    """Construct a callback that initializes a WordPiece piece encoder
    model.

    path (Path):
        Path to the serialized WordPiece model.
    """

    def load(model, X=None, Y=None):
        model.attrs["wordpiece_processor"] = WordPieceProcessor.from_file(str(path))
        model.attrs["bos_piece"] = bos_piece
        model.attrs["eos_piece"] = eos_piece
        model.attrs["unk_piece"] = unk_piece
        model.attrs["lowercase"] = lowercase
        model.attrs["strip_accents"] = strip_accents
        return model

    return load


def _bert_preprocess(token: str) -> List[str]:
    """Split a token on punctuation characters. For instance,
    'AWO-Mitarbeiter' is split into ['AWO', '-', 'Mitarbeiter']"""
    tokens = []
    in_word = False
    while token:
        char = token[0]
        token = token[1:]
        if _is_bert_punctuation(char):
            tokens.append([char])
            in_word = False
        else:
            if in_word:
                tokens[-1].append(char)
            else:
                tokens.append([char])
                in_word = True
    return ["".join(t) for t in tokens]


def _is_bert_punctuation(char: str) -> bool:
    """Checks whether `char` is a punctuation character."""
    # ASCII punctuation from HF tranformers, since we need to split
    # in the same way.
    cp = ord(char)
    if (
        (cp >= 33 and cp <= 47)
        or (cp >= 58 and cp <= 64)
        or (cp >= 91 and cp <= 96)
        or (cp >= 123 and cp <= 126)
    ):
        return True

    return unicodedata.category(char).startswith("P")


def _strip_accents(token: str) -> str:
    token = unicodedata.normalize("NFD", token)
    return "".join([char for char in token if unicodedata.category(char) != "Mn"])



import pytest
from thinc.api import Ragged, registry, get_current_ops

from curated_transformers.tokenization.bbpe_encoder import build_byte_bpe_encoder_v1
from curated_transformers.tokenization.hf_loader import build_hf_piece_encoder_loader_v1
from curated_transformers._compat import has_hf_transformers


@pytest.fixture
def toy_encoder(test_dir):
    encoder = build_byte_bpe_encoder_v1()
    encoder.init = registry.model_loaders.get("curated-transformers.ByteBPELoader.v1")(
        vocab_path=test_dir / "toy-vocab.json", merges_path=test_dir / "toy-merges.txt"
    )
    encoder.initialize()
    return encoder


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_bbpe_encoder_hf_model(sample_docs):
    encoder = build_byte_bpe_encoder_v1()
    encoder.init = build_hf_piece_encoder_loader_v1(name="roberta-base")
    encoder.initialize()
    encoding = encoder.predict(sample_docs)
    _check_roberta_base_encoder(encoding)


def test_bbpe_encoder(toy_encoder, sample_docs):
    encoding = toy_encoder.predict(sample_docs)
    _check_toy_encoder(encoding)


def test_serialize(toy_encoder):
    encoder_bytes = toy_encoder.to_bytes()
    toy_encoder2 = build_byte_bpe_encoder_v1()
    toy_encoder2.from_bytes(encoder_bytes)
    assert (
        toy_encoder.attrs["byte_bpe_processor"].vocab
        == toy_encoder2.attrs["byte_bpe_processor"].vocab
    )
    assert (
        toy_encoder.attrs["byte_bpe_processor"].merges
        == toy_encoder2.attrs["byte_bpe_processor"].merges
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_serialize_hf_model(sample_docs):
    encoder = build_byte_bpe_encoder_v1()
    encoder.init = build_hf_piece_encoder_loader_v1(name="roberta-base")
    encoder.initialize()
    encoder_bytes = encoder.to_bytes()
    encoder2 = build_byte_bpe_encoder_v1()
    encoder2.from_bytes(encoder_bytes)
    encoding = encoder2.predict(sample_docs)
    _check_roberta_base_encoder(encoding)


def _check_toy_encoder(encoding):
    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops = get_current_ops()
    ops.xp.testing.assert_array_equal(
        encoding[0].lengths, [1, 1, 1, 1, 3, 1, 1, 6, 1, 1]
    )
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd,
        [0, 44, 997, 262, 305, 334, 79, 342, 262, 388, 79, 302, 70, 472, 72, 17, 2],
    )

    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 3, 1, 1, 2, 4, 3, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd,
        [0, 55, 841, 321, 362, 579, 324, 294, 291, 494, 131, 106, 270, 307, 79, 17, 2],
    )


def _check_roberta_base_encoder(encoding):
    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops = get_current_ops()
    ops.xp.testing.assert_array_equal(
        encoding[0].lengths, [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    )
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd, [0, 100, 794, 10, 1816, 19, 10, 27608, 4, 2]
    )

    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 1, 1, 1, 1, 2, 1, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd, [0, 5625, 52, 40, 3529, 181, 48344, 5749, 4, 2]
    )



from typing import Optional
from io import BytesIO
from pathlib import Path
import spacy
from spacy import Language
from spacy.cli import app
from thinc.api import Model, PyTorchShim, pytorch_to_torchscript_wrapper
import torch
import torch.nn.quantized as nnq
from torch.nn import Embedding, Linear, Module, MSELoss
from torch.quantization import qconfig
from typer import Argument as Arg, Option
import warnings

from ..errors import Warnings, Errors
from ..pipeline.transformer import Transformer


MODULE_QUANTIZERS = {
    Embedding: qconfig.float_qparams_weight_only_qconfig,
    Linear: qconfig.default_dynamic_qconfig,
}


@app.command("quantize-transformer")
def quantize_cli(
    model_path: Path = Arg(..., help="Model to quantize", exists=True, allow_dash=True),
    output_path: Path = Arg(..., help="Output directory to store quantized model in"),
    max_mse_loss: Optional[float] = Option(
        None, "--max-mse-loss", help="Maximum MSE loss of quantized parameters"
    ),
    skip_embeds: bool = Option(
        False, "--skip-embeds", help="Do not quantize embeddings"
    ),
    skip_linear: bool = Option(
        False, "--skip-linear", help="Do not quantize linear layers"
    ),
):
    """
    Quantize a curated transformers model to reduce its size.
    """
    nlp = spacy.load(model_path)
    nlp_quantize_dynamic(
        nlp, max_mse_loss=max_mse_loss, skip_embeds=skip_embeds, skip_linear=skip_linear
    )
    nlp.to_disk(output_path)


def size_of_model(model: Module):
    filelike = BytesIO()
    torch.save(model.state_dict(), filelike)
    return filelike.tell()


def nlp_quantize_dynamic(
    nlp: Language,
    *,
    max_mse_loss: Optional[float] = None,
    skip_embeds: bool = False,
    skip_linear: bool = False,
):
    for pipe_name, pipe in nlp.components:
        # We only quantize curated transformers, other models may not
        # be able to deal with quantization and/or TorchScript.
        if not isinstance(pipe, Transformer):
            continue

        model = pipe.model.get_ref("transformer")

        if model.name == "pytorch_script":  # Probably already quantized.
            warnings.warn(Warnings.W001)
            continue

        before_size = size_of_model(pytorch_model(model))
        quantize_dynamic(
            model,
            max_mse_loss=max_mse_loss,
            skip_embeds=skip_embeds,
            skip_linear=skip_linear,
        )
        after_size = size_of_model(pytorch_model(model))
        print(
            f"Quantized model in pipe '{pipe_name}' ({before_size/2**20:.1f} MiB -> {after_size/2**20:.1f} MiB)"
        )

        pipe.model.replace_node(model, pytorch_to_torchscript_wrapper(model))
        nlp.config["components"][pipe_name]["model"]["torchscript"] = True


def quantize_dynamic(
    model: Model,
    max_mse_loss: Optional[float] = None,
    skip_embeds: bool = False,
    skip_linear: bool = False,
):
    assert model.name == "pytorch"
    assert isinstance(model.shims[0], PyTorchShim)

    pytorch_model = model.shims[0]._model
    quantize_types = {}

    if not skip_embeds:
        quantize_types[Embedding] = MODULE_QUANTIZERS[Embedding]

    if not skip_linear:
        quantize_types[Linear] = MODULE_QUANTIZERS[Linear]  # type: ignore

    quantized_model = torch.quantization.quantize_dynamic(
        pytorch_model,
        quantize_types,
        dtype=torch.qint8,
    )

    if max_mse_loss:
        quantized_model = requantize_with_max_loss(
            pytorch_model, quantized_model, max_mse_loss
        )

    model.shims[0]._model = quantized_model


def requantize_with_max_loss(
    model: Module, quantized_model: Module, max_mse_loss: float
):
    loss = MSELoss()

    modules = {name: child for name, child in model.named_modules()}
    skipped_modules = []
    module_qconfig = {}
    for name, qmodule in quantized_model.named_modules():
        if not (isinstance(qmodule, nnq.Embedding) or isinstance(qmodule, nnq.Linear)):
            continue

        module = modules[name]
        mse_loss = loss(torch.dequantize(qmodule.weight()), module.weight)
        if mse_loss <= max_mse_loss:
            module_qconfig[name] = MODULE_QUANTIZERS[type(module)]
        else:
            skipped_modules.append(name)

    print(f"Skipping modules with MSE > {max_mse_loss}: {', '.join(skipped_modules)}")

    return torch.quantization.quantize_dynamic(
        model,
        module_qconfig,
        dtype=torch.qint8,
    )


def pytorch_model(model: Model) -> Module:
    if model.name != "pytorch" or not isinstance(model.shims[0], PyTorchShim):
        raise ValueError(Errors.E001.format(model_name=model.name))

    return model.shims[0]._model



from pathlib import Path
import pytest
import spacy

from curated_transformers.tokenization.wordpiece_encoder import (
    build_wordpiece_encoder_v1,
)
from curated_transformers.util import registry


def pytest_addoption(parser):
    try:
        parser.addoption("--slow", action="store_true", help="include slow tests")
    # Options are already added, e.g. if conftest is copied in a build pipeline
    # and runs twice
    except ValueError:
        pass


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: include slow tests")


def pytest_runtest_setup(item):
    def getopt(opt):
        # When using 'pytest --pyargs spacy' to test an installed copy of
        # spacy, pytest skips running our pytest_addoption() hook. Later, when
        # we call getoption(), pytest raises an error, because it doesn't
        # recognize the option we're asking about. To avoid this, we need to
        # pass a default value. We default to False, i.e., we act like all the
        # options weren't given.
        return item.config.getoption(f"--{opt}", False)

    # Integration of boolean flags
    for opt in ["slow"]:
        if opt in item.keywords and not getopt(opt):
            pytest.skip(f"need --{opt} option to run")


@pytest.fixture
def test_dir(request):
    print(request.fspath)
    return Path(request.fspath).parent


@pytest.fixture
def sample_docs_with_spaces():
    nlp = spacy.blank("en")
    doc1 = nlp.make_doc("I saw a girl    with a telescope.")
    doc2 = nlp.make_doc("Today we    will eat    poké bowl.  ")
    return [doc1, doc2]


@pytest.fixture
def sample_docs():
    nlp = spacy.blank("en")
    doc1 = nlp.make_doc("I saw a girl with a telescope.")
    doc2 = nlp.make_doc("Today we will eat poké bowl.")
    return [doc1, doc2]


@pytest.fixture
def wordpiece_toy_model_path():
    return Path(__file__).parent / "tokenization" / "toy.wordpieces"


@pytest.fixture
def wordpiece_toy_encoder(wordpiece_toy_model_path):
    encoder = build_wordpiece_encoder_v1()
    encoder.init = registry.model_loaders.get(
        "curated-transformers.WordpieceLoader.v1"
    )(path=wordpiece_toy_model_path)
    encoder.initialize()
    return encoder



from typing import List, Tuple, Callable, cast
from thinc.layers.pytorchwrapper import PyTorchWrapper_v2
from thinc.model import Model
from thinc.types import ArgsKwargs, Ragged
from thinc.shims.pytorch_grad_scaler import PyTorchGradScaler
from thinc.util import torch2xp, xp2torch
from torch import Tensor
from spacy.util import SimpleFrozenDict

from ..util import all_equal
from .types import ScalarWeightInT, ScalarWeightOutT, ScalarWeightModelT
from .pytorch.scalar_weight import ScalarWeight
from ..errors import Errors


def build_scalar_weight_v1(
    *,
    num_layers: int,
    dropout_prob: float = 0.1,
    mixed_precision: bool = False,
    grad_scaler_config: dict = SimpleFrozenDict(),
) -> ScalarWeightModelT:
    """Construct a model that accepts a list of transformer layer
    outputs and returns a weighted representation of the same.

    num_layers (int):
        Number of transformer hidden layers.
    dropout_prob (float):
        Dropout probability.
    mixed_precision (bool):
        Use mixed-precision training.
    grad_scaler_config (dict):
        Configuration passed to the PyTorch gradient scaler.
    """
    if isinstance(grad_scaler_config, SimpleFrozenDict):
        # Create a new, mutable dict instance.
        grad_scaler_config = {}

    if "enabled" not in grad_scaler_config:
        grad_scaler_config["enabled"] = mixed_precision

    # Increment number of layers by one to include the embedding layer.
    scalar_weighting_layer = ScalarWeight(
        num_layers=num_layers + 1, dropout_prob=dropout_prob
    )

    model = PyTorchWrapper_v2(
        scalar_weighting_layer,
        convert_inputs=_convert_inputs,
        convert_outputs=_convert_outputs,
        mixed_precision=mixed_precision,
        grad_scaler=PyTorchGradScaler(**grad_scaler_config),
    )

    return model


def _convert_inputs(
    model: Model, X: ScalarWeightInT, is_train: bool = False
) -> Tuple[ArgsKwargs, Callable[[ArgsKwargs], List[Ragged]]]:
    ops = model.ops

    batch_size = len(X)
    seq_lens = [x[0].data.shape[0] for x in X]
    max_seq_len = max(seq_lens)
    num_layers = [len(x) for x in X]
    if not all_equal(num_layers):
        raise ValueError(Errors.E015.format(layer_counts=set(num_layers)))
    layer_widths = [layer.data.shape[1] for x in X for layer in x]
    if not all_equal(layer_widths):
        raise ValueError(Errors.E016.format(hidden_widths=set(layer_widths)))

    # [batch_size, max_seq_len, num_layers, layer_width]
    Xops = ops.alloc4f(batch_size, max_seq_len, num_layers[0], layer_widths[0])
    for doc_idx, layers in enumerate(X):
        seq_len = seq_lens[doc_idx]
        for layer_idx, data in enumerate(layers):
            Xops[doc_idx, :seq_len, layer_idx, :] = data.dataXd  # type: ignore
    Xt = xp2torch(Xops, requires_grad=True)

    def convert_from_torch_backward(d_inputs: ArgsKwargs):
        # [batch, seq, num_layers, layer_width]
        dt_inputs: Tensor = cast(Tensor, d_inputs.args[0])

        dX = []
        for doc_idx in range(dt_inputs.shape[0]):
            seq_len = seq_lens[doc_idx]
            lengths = X[doc_idx][0].lengths
            dX_layers = []
            for layer_idx in range(dt_inputs.shape[2]):
                dX_layer = dt_inputs[doc_idx, :seq_len, layer_idx, :]
                dX_layers.append(
                    Ragged(data=torch2xp(dX_layer, ops=ops), lengths=lengths)
                )
            dX.append(dX_layers)
        return dX

    output = ArgsKwargs(args=(Xt,), kwargs={})
    return output, convert_from_torch_backward


def _convert_outputs(
    model: Model, inputs_outputs: Tuple[ScalarWeightInT, Tensor], is_train: bool
) -> Tuple[ScalarWeightOutT, Callable[[List[Ragged]], ArgsKwargs]]:
    ops = model.ops
    X, Yt = inputs_outputs

    Y = []
    for doc_idx in range(len(X)):
        seq_len = X[doc_idx][0].data.shape[0]
        lengths = X[doc_idx][0].lengths
        # [batch, seq, width]
        Y_layer = Yt[doc_idx, :seq_len, :]  # type: ignore
        Y.append(Ragged(torch2xp(Y_layer, ops=model.ops), lengths=lengths))

    def convert_for_torch_backward(dY: List[Ragged]) -> ArgsKwargs:
        max_seq_len = max(y.data.shape[0] for y in dY)
        width = dY[0].data.shape[1]

        dYt_ops = ops.alloc3f(len(dY), max_seq_len, width)
        for doc_idx in range(len(dY)):
            seq_len = dY[doc_idx].data.shape[0]
            dYt_ops[doc_idx, :seq_len, :] = dY[doc_idx].dataXd  # type: ignore

        dYt = xp2torch(dYt_ops)
        return ArgsKwargs(
            args=(Yt,),
            kwargs={"grad_tensors": dYt},
        )

    return Y, convert_for_torch_backward



from typing import List
from thinc.api import Model, chain
from thinc.types import Floats2d, Ragged

from curated_transformers.models.output import TransformerModelOutput
from curated_transformers.models.remove_eos_bos import remove_bos_eos
from curated_transformers.models.with_non_ws_tokens import with_non_ws_tokens


def _mock_transformer() -> Model[List[Floats2d], TransformerModelOutput]:
    def forward(model: Model, X: List[Floats2d], is_train: bool):
        def backprop(dY):
            model.attrs["last_dY"] = dY
            return dY

        return (
            TransformerModelOutput(
                outputs=[
                    [
                        Ragged(
                            model.ops.xp.ones((x.dataXd.shape[0], 2)), lengths=x.lengths
                        )
                    ]
                    for x in X
                ],
                last_layer_only=False,
            ),
            backprop,
        )

    return Model("mock_transformer", forward)


def test_with_non_ws_tokens(sample_docs_with_spaces, wordpiece_toy_encoder):
    mock_transformer = _mock_transformer()
    model = with_non_ws_tokens(
        chain(wordpiece_toy_encoder, mock_transformer, remove_bos_eos())
    )
    model.initialize()
    Y, backprop = model(sample_docs_with_spaces, is_train=True)

    yl0 = Y.all_outputs[0][0]
    y1_check = model.ops.xp.ones((15, 2))
    y1_check[6, :] = 0.0
    model.ops.xp.testing.assert_array_equal(yl0.dataXd, y1_check)

    yl1 = Y.all_outputs[1][0]
    y2_check = model.ops.xp.ones((17, 2))
    y2_check[4, :] = 0.0
    y2_check[8, :] = 0.0
    y2_check[16, :] = 0.0
    model.ops.xp.testing.assert_array_equal(yl1.dataXd, y2_check)

    dY = [
        [
            Ragged(
                model.ops.xp.arange(float(yl0.dataXd.size)).reshape(yl0.dataXd.shape),
                lengths=yl0.lengths,
            )
        ],
        [
            Ragged(
                model.ops.xp.arange(float(yl1.dataXd.size)).reshape(yl1.dataXd.shape),
                lengths=yl1.lengths,
            )
        ],
    ]

    backprop(dY)

    transformer_dY = mock_transformer.attrs["last_dY"]
    model.ops.xp.testing.assert_array_equal(
        transformer_dY[0][0].dataXd,
        [
            [0.0, 0.0],
            [0.0, 1.0],
            [2.0, 3.0],
            [4.0, 5.0],
            [6.0, 7.0],
            [8.0, 9.0],
            [10.0, 11.0],
            [14.0, 15.0],
            [16.0, 17.0],
            [18.0, 19.0],
            [20.0, 21.0],
            [22.0, 23.0],
            [24.0, 25.0],
            [26.0, 27.0],
            [28.0, 29.0],
            [0.0, 0.0],
        ],
    )
    model.ops.xp.testing.assert_array_equal(
        transformer_dY[1][0].dataXd,
        [
            [0.0, 0.0],
            [0.0, 1.0],
            [2.0, 3.0],
            [4.0, 5.0],
            [6.0, 7.0],
            [10.0, 11.0],
            [12.0, 13.0],
            [14.0, 15.0],
            [18.0, 19.0],
            [20.0, 21.0],
            [22.0, 23.0],
            [24.0, 25.0],
            [26.0, 27.0],
            [28.0, 29.0],
            [30.0, 31.0],
            [0.0, 0.0],
        ],
    )



from typing import Any, Callable, Any, Iterable, Dict, TYPE_CHECKING
from functools import partial
from spacy.language import Language
import itertools
import thinc

from .errors import Errors

if TYPE_CHECKING:
    from .pipeline.transformer import Transformer

thinc.registry.create("model_loaders", entry_points=True)
registry = thinc.registry


def all_equal(iterable: Iterable[Any]) -> bool:
    """Return True if all the elements are equal to each other
    (or if the input is an empty sequence), False otherwise."""
    g = itertools.groupby(iterable)
    return next(g, True) and not next(g, False)  # type: ignore


def gradual_transformer_unfreezing_per_pipe(
    nlp: Language,
    callback_args: Dict[str, Any],
    freeze_params: Dict[str, int],
):
    current_step = callback_args["step"]

    # Scoped import to avoid import cycles.
    from .pipeline.transformer import Transformer

    for name, pipe in nlp.components:
        unfreeze_step = freeze_params.get(name)
        if unfreeze_step is None:
            continue
        elif not isinstance(pipe, Transformer):
            raise TypeError(Errors.E025.format(pipe_name=name))

        pipe.frozen = current_step < unfreeze_step


def gradual_transformer_unfreezing_all_pipes(
    nlp: Language, callback_args: Dict[str, Any], unfreeze_step: int
):
    current_step = callback_args["step"]

    # Scoped import to avoid import cycles.
    from .pipeline.transformer import Transformer

    for _, pipe in nlp.components:
        if not isinstance(pipe, Transformer):
            continue

        pipe.frozen = current_step < unfreeze_step


def create_gradual_transformer_unfreezing(
    target_pipes: Dict[str, int]
) -> Callable[[Language, Dict[str, Any]], None]:
    """Construct a callback that can be used to gradually unfreeze the
    weights of one or more Transformer components during training. This
    can be used to prevent catastrophic forgetting during fine-tuning.

    target_pipes (Dict[str, int]):
        A dictionary whose keys and values correspond to the names of Transformer
        components and the training step at which they should be unfrozen respectively.
    """
    unfreeze_step_all_pipes = target_pipes.get("*")
    if unfreeze_step_all_pipes is not None and len(target_pipes) > 1:
        raise ValueError(Errors.E013)

    if unfreeze_step_all_pipes is not None:
        return partial(
            gradual_transformer_unfreezing_all_pipes,
            unfreeze_step=unfreeze_step_all_pipes,
        )
    else:
        return partial(
            gradual_transformer_unfreezing_per_pipe,
            freeze_params=target_pipes,
            last_unfreeze_step=max(target_pipes.values()),
        )



import pytest
from spacy import util
from spacy.training import Example
from thinc.api import Config, reduce_mean

from curated_transformers.models.listeners import build_transformer_layers_listener_v1
from curated_transformers._compat import has_hf_transformers

cfg_string_transformer_layers_listener = """
    # TransformerLayersListener

    [nlp]
    lang = "en"
    pipeline = ["transformer"]

    [components]

    [components.transformer]
    factory = "curated_transformer"
    all_layer_outputs = True

    [components.transformer.model]
    @architectures = "curated-transformers.BertTransformer.v1"
    vocab_size = 28996
    num_hidden_layers = 2
    hidden_width = 60
    piece_encoder = {"@architectures":"curated-transformers.BertWordpieceEncoder.v1"}
    with_spans = {"@architectures":"curated-transformers.WithStridedSpans.v1"}

    [initialize]

    [initialize.components]

    [initialize.components.transformer]

    [initialize.components.transformer.piecer_loader]
    @model_loaders = "curated-transformers.HFPieceEncoderLoader.v1"
    name = "bert-base-cased"
"""


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_transformer_layers_listener():
    config = Config().from_str(cfg_string_transformer_layers_listener)
    nlp = util.load_model_from_config(config, auto_fill=True, validate=True)

    transformer = nlp.get_pipe("transformer")
    listener = build_transformer_layers_listener_v1(
        layers=2, width=60, pooling=reduce_mean()
    )
    transformer.add_listener(listener, "test")

    docs = [
        nlp.make_doc("Let's test a transformer."),
        nlp.make_doc("Can it handle more than one doc?"),
    ]
    examples = [Example.from_dict(doc, {}) for doc in docs]

    nlp.initialize(lambda: examples)

    # Check prediction.
    transformer.pipe(docs)
    Y = listener.predict(docs)

    hidden_size = 60
    layers = 3  # embed + hidden layers

    assert len(Y) == 2
    assert len(Y[0]) == layers
    assert len(Y[1]) == layers
    assert Y[0][-1].shape == (len(docs[0]), hidden_size)
    assert Y[1][-1].shape == (len(docs[1]), hidden_size)

    # Check update.

    transformer.update(examples)
    Y, backprop = listener.begin_update(docs)
    assert len(Y) == 2
    assert len(Y[0]) == layers
    assert len(Y[1]) == layers
    assert Y[0][-1].shape == (len(docs[0]), hidden_size)
    assert Y[1][-1].shape == (len(docs[1]), hidden_size)

    dY_doc0 = listener.ops.alloc2f(len(docs[0]), 60)
    dY_doc1 = listener.ops.alloc2f(len(docs[1]), 60)

    backprop([layers * [dY_doc0], layers * [dY_doc1]])



from typing import Dict, OrderedDict
import torch
import re


from .albert.encoder import AlbertEncoder
from .bert.encoder import BertEncoder
from .curated_transformer import CuratedTransformer, CuratedEncoderT
from .roberta.encoder import RobertaEncoder
from ..._compat import transformers
from ...errors import Errors

SUPPORTED_MODEL_TYPES = ["albert", "bert", "camembert", "roberta", "xlm-roberta"]
SUPPORTED_CURATED_ENCODERS = (AlbertEncoder, BertEncoder, RobertaEncoder)


def _check_supported_hf_models(model_type: str):
    if model_type not in SUPPORTED_MODEL_TYPES:
        raise ValueError(
            Errors.E007.format(
                unsupported_model=model_type, supported_models=SUPPORTED_MODEL_TYPES
            )
        )


def convert_pretrained_model_for_encoder(
    transformer: CuratedTransformer[CuratedEncoderT],
    params: OrderedDict[str, torch.Tensor],
) -> Dict[str, torch.Tensor]:
    """Converts parameters from a compatible pre-trained model to
    parameters that can be consumed by the given curated transformer.

    Returns the state_dict that can be directly loaded by the curated
    transformer.
    """
    params = _rename_old_hf_names(params)
    encoder = transformer.curated_encoder

    if isinstance(encoder, AlbertEncoder):
        converted = _convert_albert_base_state(params)
    elif isinstance(encoder, BertEncoder):
        converted = _convert_bert_base_state(params)
    elif isinstance(encoder, RobertaEncoder):
        converted = _convert_roberta_base_state(params)
    else:
        raise TypeError(
            Errors.E026.format(
                unsupported_encoder=type(encoder),
                supported_encoders=SUPPORTED_CURATED_ENCODERS,
            )
        )

    return _add_curated_encoder_prefix(converted)


def convert_hf_pretrained_model_parameters(
    hf_model: "transformers.PreTrainedModel",
) -> Dict[str, torch.Tensor]:
    """Converts HF model parameters to parameters that can be consumed by
    our implementation of the Transformer.

    Returns the state_dict that can be directly loaded by one of the
    curated transformers.
    """
    _check_supported_hf_models(hf_model.config.model_type)  # type: ignore

    converters = {
        "albert": _convert_albert_base_state,
        "bert": _convert_bert_base_state,
        "camembert": _convert_roberta_base_state,
        "roberta": _convert_roberta_base_state,
        "xlm-roberta": _convert_roberta_base_state,
    }

    converted = converters[hf_model.config.model_type](hf_model.state_dict())  # type: ignore

    return _add_curated_encoder_prefix(converted)


def _add_curated_encoder_prefix(
    converted: Dict[str, torch.Tensor]
) -> Dict[str, torch.Tensor]:
    return {f"curated_encoder.{k}": v for k, v in converted.items()}


def _rename_old_hf_names(
    params: OrderedDict[str, torch.Tensor],
) -> OrderedDict[str, torch.Tensor]:
    out = OrderedDict()
    for name, parameter in params.items():
        name = re.sub(r"\.gamma$", ".weight", name)
        name = re.sub(r"\.beta$", ".bias", name)
        out[name] = parameter
    return out


def _convert_albert_base_state(
    params: OrderedDict[str, torch.Tensor],
) -> Dict[str, torch.Tensor]:
    # Strip the `albert` prefix from ALBERT model parameters.
    stripped_params = {re.sub(r"^albert\.", "", k): v for k, v in params.items()}

    # The ALBERT encoder parameters have the following form:
    #
    # encoder.albert_layer_groups.{hidden_group}.albert_layers.{inner_layer}.{param_name}
    #
    # hidden_group is in [0, num_hidden_group)
    # inner_layer is in [0, inner_group_num)

    out = {}
    for name, parameter in stripped_params.items():
        if "encoder.albert_layer" not in name:
            continue

        # TODO: Make these substitutions less ugly.

        # Remove the prefix and rename.
        name = re.sub(r"^encoder\.", "", name)

        # Layer groups
        name = re.sub(r"^albert_layer_groups\.", "groups.", name)

        # Inner layers.
        name = re.sub(r"\.albert_layers\.", ".group_layers.", name)

        # Attention blocks.
        name = re.sub(r"\.attention\.", ".mha.", name)
        name = re.sub(r"\.mha\.LayerNorm", r".attn_output_layernorm", name)
        name = re.sub(r"\.mha\.dense\.", r".mha.output.", name)

        # Pointwise feed-forward layers.
        name = re.sub(r"\.ffn\.", r".ffn.intermediate.", name)
        name = re.sub(r"\.ffn_output\.", r".ffn.output.", name)
        name = re.sub(
            r"\.full_layer_layer_norm\.",
            r".ffn_output_layernorm.",
            name,
        )

        out[name] = parameter

    # Rename and move embedding parameters to the inner BertEmbeddings module.
    out["embeddings.word_embeddings.weight"] = stripped_params[
        "embeddings.word_embeddings.weight"
    ]
    out["embeddings.token_type_embeddings.weight"] = stripped_params[
        "embeddings.token_type_embeddings.weight"
    ]
    out["embeddings.position_embeddings.weight"] = stripped_params[
        "embeddings.position_embeddings.weight"
    ]
    out["embeddings.layer_norm.weight"] = stripped_params["embeddings.LayerNorm.weight"]
    out["embeddings.layer_norm.bias"] = stripped_params["embeddings.LayerNorm.bias"]

    # Embedding projection
    out["embeddings.projection.weight"] = stripped_params[
        "encoder.embedding_hidden_mapping_in.weight"
    ]
    out["embeddings.projection.bias"] = stripped_params[
        "encoder.embedding_hidden_mapping_in.bias"
    ]

    return _merge_qkv_albert(out)


def _convert_bert_base_state(
    params: OrderedDict[str, torch.Tensor],
) -> Dict[str, torch.Tensor]:
    out = {}

    # Strip the `bert` prefix from BERT model parameters.
    stripped_params = {re.sub(r"^bert\.", "", k): v for k, v in params.items()}

    for name, parameter in stripped_params.items():
        if "encoder.layer." not in name:
            continue

        # TODO: Make these substitutions less ugly.

        # Remove the prefix and rename the internal 'layers' variable.
        name = re.sub(r"^encoder\.", "", name)
        name = re.sub(r"^layer", "layers", name)

        # The HF model has one more level of indirection for the output layers in their
        # attention heads and the feed-forward network layers.
        name = re.sub(r"\.attention\.self\.(query|key|value)", r".mha.\1", name)
        name = re.sub(r"\.attention\.(output)\.dense", r".mha.\1", name)
        name = re.sub(
            r"\.attention\.output\.LayerNorm", r".attn_output_layernorm", name
        )
        name = re.sub(r"\.(intermediate)\.dense", r".ffn.\1", name)
        name = re.sub(r"(\.\d+)\.output\.LayerNorm", r"\1.ffn_output_layernorm", name)
        name = re.sub(r"(\.\d+)\.(output)\.dense", r"\1.ffn.\2", name)

        out[name] = parameter

    # Rename and move embedding parameters to the inner BertEmbeddings module.
    out["embeddings.word_embeddings.weight"] = stripped_params[
        "embeddings.word_embeddings.weight"
    ]
    out["embeddings.token_type_embeddings.weight"] = stripped_params[
        "embeddings.token_type_embeddings.weight"
    ]
    out["embeddings.position_embeddings.weight"] = stripped_params[
        "embeddings.position_embeddings.weight"
    ]
    out["embeddings.layer_norm.weight"] = stripped_params["embeddings.LayerNorm.weight"]
    out["embeddings.layer_norm.bias"] = stripped_params["embeddings.LayerNorm.bias"]

    return _merge_qkv(out)


def _convert_roberta_base_state(
    params: OrderedDict[str, torch.Tensor],
) -> Dict[str, torch.Tensor]:
    out = {}

    # Strip the `roberta` prefix from XLM-Roberta model parameters.
    stripped_params = {re.sub(r"^roberta\.", "", k): v for k, v in params.items()}

    for name, parameter in stripped_params.items():
        if "encoder.layer." not in name:
            continue

        # TODO: Make these substitutions less ugly.

        # Remove the prefix and rename the internal 'layers' variable.
        name = re.sub(r"^encoder\.", "", name)
        name = re.sub(r"^layer", "layers", name)

        # The HF model has one more level of indirection for the output layers in their
        # attention heads and the feed-forward network layers.
        name = re.sub(r"\.attention\.self\.(query|key|value)", r".mha.\1", name)
        name = re.sub(r"\.attention\.(output)\.dense", r".mha.\1", name)
        name = re.sub(
            r"\.attention\.output\.LayerNorm", r".attn_output_layernorm", name
        )
        name = re.sub(r"\.(intermediate)\.dense", r".ffn.\1", name)
        name = re.sub(r"(\.\d+)\.output\.LayerNorm", r"\1.ffn_output_layernorm", name)
        name = re.sub(r"(\.\d+)\.(output)\.dense", r"\1.ffn.\2", name)

        out[name] = parameter

    # Rename and move embedding parameters to the inner BertEmbeddings module.
    out["embeddings.inner.word_embeddings.weight"] = stripped_params[
        "embeddings.word_embeddings.weight"
    ]
    out["embeddings.inner.token_type_embeddings.weight"] = stripped_params[
        "embeddings.token_type_embeddings.weight"
    ]
    out["embeddings.inner.position_embeddings.weight"] = stripped_params[
        "embeddings.position_embeddings.weight"
    ]
    out["embeddings.inner.layer_norm.weight"] = stripped_params[
        "embeddings.LayerNorm.weight"
    ]
    out["embeddings.inner.layer_norm.bias"] = stripped_params[
        "embeddings.LayerNorm.bias"
    ]

    return _merge_qkv(out)


def _merge_qkv(params: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
    out = {}
    for name, parameter in params.items():
        m = re.match(
            r"layers\.(?P<layer>[0-9]+)\.mha\.(query|key|value).(?P<param_type>weight|bias)",
            name,
        )
        if m:
            if "query" in name:
                base = f"layers.{m['layer']}.mha"
                out[f"{base}.input.{m['param_type']}"] = torch.cat(
                    [
                        parameter,
                        params[f"{base}.key.{m['param_type']}"],
                        params[f"{base}.value.{m['param_type']}"],
                    ]
                )
            continue
        out[name] = parameter

    return out


def _merge_qkv_albert(params: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
    out = {}
    for name, parameter in params.items():
        m = re.match(
            r"groups\.(?P<group>[0-9]+)\.group_layers\.(?P<layer>[0-9]+)\.mha\.(query|key|value).(?P<param_type>weight|bias)",
            name,
        )
        if m:
            if "query" in name:
                base = f"groups.{m['group']}.group_layers.{m['layer']}.mha"
                out[f"{base}.input.{m['param_type']}"] = torch.cat(
                    [
                        parameter,
                        params[f"{base}.key.{m['param_type']}"],
                        params[f"{base}.value.{m['param_type']}"],
                    ]
                )
            continue
        out[name] = parameter

    return out



from spacy.lang.ja import try_sudachi_import

try:
    import transformers

    has_hf_transformers = True
except ImportError:
    transformers = None  # type: ignore
    has_hf_transformers = False

try:
    import huggingface_hub

    has_huggingface_hub = True
except ImportError:
    huggingface_hub = None  # type: ignore
    has_huggingface_hub = False


try:
    import fugashi

    has_fugashi = True
except ImportError:
    fugashi = None
    has_fugashi = False

try:
    try_sudachi_import()
    has_sudachi = True
except ImportError:
    has_sudachi = False



from .bbpe_encoder import build_byte_bpe_encoder_loader_v1, build_byte_bpe_encoder_v1
from .char_encoder import build_char_encoder_loader_v1, build_char_encoder_v1
from .hf_loader import build_hf_piece_encoder_loader_v1
from .sentencepiece_encoder import (
    build_camembert_sentencepiece_encoder_v1,
    build_sentencepiece_encoder_loader_v1,
    build_sentencepiece_encoder_v1,
    build_xlmr_sentencepiece_encoder_v1,
)
from .wordpiece_encoder import (
    build_bert_wordpiece_encoder_v1,
    build_wordpiece_encoder_loader_v1,
    build_wordpiece_encoder_v1,
)



from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Tuple,
)

from spacy import Errors as SpacyErrors
from spacy.tokens import Doc
from thinc.model import Model
from thinc.types import Ragged, Floats2d

from .output import TransformerModelOutput
from .pooling import with_ragged_layers, with_ragged_last_layer
from .types import (
    WithRaggedLayersModelT,
    WithRaggedLastLayerModelT,
    PoolingModelT,
    ScalarWeightOutT,
    ScalarWeightModelT,
)
from .output import TransformerModelOutput
from .pooling import with_ragged_layers, with_ragged_last_layer
from .types import (
    WithRaggedLayersModelT,
    WithRaggedLastLayerModelT,
    PoolingModelT,
)
from ..errors import Errors


def build_transformer_layers_listener_v1(
    layers: int,
    width: int,
    pooling: PoolingModelT,
    upstream: str = "*",
    grad_factor: float = 1.0,
) -> Model[List[Doc], List[Floats2d]]:
    """Construct a listener layer that communicates with one or more upstream Transformer
    components. This layer extracts the output of all transformer layers and performs
    pooling over the individual pieces of each Doc token, returning their corresponding
    representations.

    layers (int):
        The the number of layers produced by the upstream transformer component,
        excluding the embedding layer.
    width (int):
        The width of the vectors produced by the upstream transformer component.
    pooling (PoolingModelT):
        Model that is used to perform pooling over the piece representations.
    upstream_name (str):
        A string to identify the 'upstream' Transformer component
        to communicate with. The upstream name should either be the wildcard
        string '*', or the name of the Transformer component.

        In almost all cases, the wildcard string will suffice as there'll only be one
        upstream Transformer component. But in certain situations, e.g: you have disjoint
        datasets for certain tasks, or you'd like to use a pre-trained pipeline but a
        downstream task requires its own token representations, you could end up with
        more than one Transformer component in the pipeline.
    grad_factor (float):
        Factor to multiply gradients with.
    """
    transformer = TransformerLayersListener(
        upstream_name=upstream,
        pooling=pooling,
        layers=layers,
        width=width,
        grad_factor=grad_factor,
    )
    return transformer


def build_last_transformer_layer_listener_v1(
    width: int,
    pooling: PoolingModelT,
    upstream: str = "*",
    grad_factor: float = 1.0,
) -> Model[List[Doc], List[Floats2d]]:
    """Construct a listener layer that communicates with one or more upstream Transformer
    components. This layer extracts the output of the last transformer layer and performs
    pooling over the individual pieces of each Doc token, returning their corresponding
    representations.

    width (int):
        The width of the vectors produced by the upstream transformer component.
    pooling (Model):
        Model that is used to perform pooling over the piece representations.
    upstream_name (str):
        A string to identify the 'upstream' Transformer component
        to communicate with. The upstream name should either be the wildcard
        string '*', or the name of the Transformer component.

        In almost all cases, the wildcard string will suffice as there'll only be one
        upstream Transformer component. But in certain situations, e.g: you have disjoint
        datasets for certain tasks, or you'd like to use a pre-trained pipeline but a
        downstream task requires its own token representations, you could end up with
        more than one Transformer component in the pipeline.
    grad_factor (float):
        Factor to multiply gradients with.
    """
    transformer = LastTransformerLayerListener(
        upstream_name=upstream, pooling=pooling, width=width, grad_factor=grad_factor
    )
    return transformer


def build_scalar_weighting_listener_v1(
    width: int,
    weighting: ScalarWeightModelT,
    pooling: PoolingModelT,
    upstream: str = "*",
    grad_factor: float = 1.0,
) -> Model[List[Doc], List[Floats2d]]:
    """Construct a listener layer that communicates with one or more upstream Transformer
    components. This layer calculates a weighted representation of all transformer layer
    outputs and performs pooling over the individual pieces of each Doc token, returning
    their corresponding representations.

    Requires its upstream Transformer components to return all layer outputs from
    their models.

    width (int):
        The width of the vectors produced by the upstream transformer component.
    weighting (Model):
        Model that is used to perform the weighting of the different layer outputs.
    pooling (Model):
        Model that is used to perform pooling over the piece representations.
    upstream_name (str):
        A string to identify the 'upstream' Transformer component
        to communicate with. The upstream name should either be the wildcard
        string '*', or the name of the Transformer component.

        In almost all cases, the wildcard string will suffice as there'll only be one
        upstream Transformer component. But in certain situations, e.g: you have disjoint
        datasets for certain tasks, or you'd like to use a pre-trained pipeline but a
        downstream task requires its own token representations, you could end up with
        more than one Transformer component in the pipeline.
    grad_factor (float):
        Factor to multiply gradients with.
    """
    transformer = ScalarWeightingListener(
        upstream_name=upstream,
        weighting=weighting,
        pooling=pooling,
        width=width,
        grad_factor=grad_factor,
    )
    return transformer


class TransformerListener(Model):
    """A layer that gets fed its answers from an upstream Transformer component.

    The TransformerListener layer is used as a sublayer within a component such
    as a parser, NER or text categorizer. Usually you'll have multiple listeners
    connecting to a single upstream Transformer component that's earlier in the
    pipeline. These layers act as proxies, passing the predictions
    from the Transformer component into downstream components and communicating
    gradients back upstream.
    """

    upstream_name: str
    _batch_id: Optional[int]
    _outputs: Optional[TransformerModelOutput]
    _backprop: Optional[Callable[[List[List[Ragged]], Tuple[int]], Any]]

    @classmethod
    def get_batch_id(cls, inputs: Iterable[Doc]) -> int:
        """Calculate a content-sensitive hash of the batch of documents, to check
        whether the next batch of documents is unexpected.
        """
        return sum(sum(token.orth for token in doc) for doc in inputs)

    def receive(
        self,
        batch_id: int,
        outputs: TransformerModelOutput,
        backprop: Callable[[List[List[Ragged]], Tuple[int]], Any],
    ) -> None:
        """Store a batch of training predictions and a backprop callback. The
        predictions and callback are produced by the upstream Transformer component,
        and later will be used when the listener's component's model is called.
        """
        self._batch_id = batch_id
        self._outputs = outputs
        self._backprop = backprop

    def verify_inputs(self, inputs: Iterable[Doc]) -> bool:
        """Check that the batch of Doc objects matches the ones we have a
        prediction for.
        """
        if self._batch_id is None and self._outputs is None:
            raise ValueError(SpacyErrors.E954)
        else:
            batch_id = self.get_batch_id(inputs)
            if batch_id != self._batch_id:
                raise ValueError(
                    SpacyErrors.E953.format(id1=batch_id, id2=self._batch_id)
                )
            else:
                return True


class TransformerLayersListener(TransformerListener):
    """Passes through the pooled representations of all transformer layers.

    Requires its upstream Transformer component to return all layer outputs from
    its model.
    """

    name = "transformer_layers_listener"

    def __init__(
        self,
        upstream_name: str,
        pooling: PoolingModelT,
        width: int,
        layers: int,
        grad_factor: float,
    ) -> None:
        """
        upstream_name (str):
            A string to identify the 'upstream' Transformer component
            to communicate with. The upstream name should either be the wildcard
            string '*', or the name of the Transformer component.

            In almost all cases, the wildcard string will suffice as there'll only be one
            upstream Transformer component. But in certain situations, e.g: you have disjoint
            datasets for certain tasks, or you'd like to use a pre-trained pipeline but a
            downstream task requires its own token representations, you could end up with
            more than one Transformer component in the pipeline.
        pooling (PoolingModelT):
            Model that is used to perform pooling over the piece representations.
        width (int):
            The width of the vectors produced by the upstream transformer component.
        layers (int):
            The the number of layers produced by the upstream transformer component,
            excluding the embedding layer.
        grad_factor (float):
            Factor to multiply gradients with.
        """
        Model.__init__(
            self,
            name=self.name,
            forward=tranformer_layers_listener_forward,
            dims={"nO": width},
            layers=[with_ragged_layers(pooling)],
            attrs={
                "grad_factor": grad_factor,
                "layers": layers + 1,
            },
            refs={"pooling": pooling},
        )
        self.upstream_name = upstream_name
        self._batch_id = None
        self._outputs = None
        self._backprop = None


def tranformer_layers_listener_forward(
    model: TransformerLayersListener, docs: Iterable[Doc], is_train: bool
) -> Tuple[List[List[Floats2d]], Callable[[Any], Any]]:
    pooling: WithRaggedLayersModelT = model.layers[0]
    grad_factor: float = model.attrs["grad_factor"]
    n_layers: int = model.attrs["layers"]

    if is_train:
        assert model._outputs is not None
        if model._outputs.last_layer_only:
            raise ValueError(
                Errors.E012.format(listener_name="TransformerLayersListener")
            )

        model.verify_inputs(docs)

        Y, backprop_pooling = pooling(model._outputs.all_outputs, is_train)

        def backprop(dY):
            dX = backprop_pooling(dY)

            if grad_factor != 1.0:
                for dX_doc in dX:
                    for dX_layer in dX_doc:
                        dX_layer.data *= grad_factor

            outputs_to_backprop = tuple(i for i in range(0, model._outputs.num_outputs))
            dX = model._backprop(dX, outputs_to_backprop=outputs_to_backprop)

            model._batch_id = None
            model._outputs = None
            model._backprop = None

            return dX

        return Y, backprop
    else:
        width = model.get_dim("nO")

        no_trf_data = [doc._.trf_data is None for doc in docs]
        if any(no_trf_data):
            assert all(no_trf_data)
            return [
                [model.ops.alloc2f(len(doc), width) for _ in range(n_layers)]
                for doc in docs
            ], lambda dY: []

        if any(doc._.trf_data.last_layer_only for doc in docs):
            raise ValueError(
                Errors.E012.format(listener_name="TransformerLayersListener")
            )

        return pooling.predict(docs), lambda dY: []


class LastTransformerLayerListener(TransformerListener):
    """Extracts the output of the last transformer layer and performs pooling over the
    individual pieces of each Doc token, returning their corresponding representations.
    """

    name = "last_transformer_layer_listener"

    def __init__(
        self,
        upstream_name: str,
        pooling: PoolingModelT,
        width: int,
        grad_factor: float,
    ) -> None:
        """
        upstream_name (str):
            A string to identify the 'upstream' Transformer component
            to communicate with. The upstream name should either be the wildcard
            string '*', or the name of the Transformer component.

            In almost all cases, the wildcard string will suffice as there'll only be one
            upstream Transformer component. But in certain situations, e.g: you have disjoint
            datasets for certain tasks, or you'd like to use a pre-trained pipeline but a
            downstream task requires its own token representations, you could end up with
            more than one Transformer component in the pipeline.
        width (int):
            The width of the vectors produced by the upstream transformer component.
        pooling (Model):
            Model that is used to perform pooling over the piece representations.
        grad_factor (float):
            Factor to multiply gradients with.
        """
        Model.__init__(
            self,
            name=self.name,
            forward=last_transformer_layer_listener_forward,
            dims={"nO": width},
            layers=[with_ragged_last_layer(pooling)],
            attrs={"grad_factor": grad_factor},
            refs={"pooling": pooling},
        )
        self.upstream_name = upstream_name
        self._batch_id = None
        self._outputs = None
        self._backprop = None


def last_transformer_layer_listener_forward(
    model: LastTransformerLayerListener, docs: Iterable[Doc], is_train: bool
) -> Tuple[List[Floats2d], Callable[[Any], Any]]:
    pooling: WithRaggedLastLayerModelT = model.layers[0]
    grad_factor: float = model.attrs["grad_factor"]

    if is_train:
        model.verify_inputs(docs)
        assert model._outputs is not None
        Y, backprop_pooling = pooling(model._outputs.last_hidden_layer_states, is_train)

        def backprop(dY):
            dX_pooling = backprop_pooling(dY)
            if grad_factor != 1.0:
                for dx in dX_pooling:
                    dx.data *= grad_factor
            dX = model._backprop([[d] for d in dX_pooling], outputs_to_backprop=(-1,))
            model._batch_id = None
            model._outputs = None
            model._backprop = None
            return dX

        return Y, backprop
    else:
        width = model.get_dim("nO")

        no_trf_data = [doc._.trf_data is None for doc in docs]
        if any(no_trf_data):
            assert all(no_trf_data)
            return [model.ops.alloc2f(len(doc), width) for doc in docs], lambda dY: []

        return pooling.predict(docs), lambda dY: []


class ScalarWeightingListener(TransformerListener):
    """Calculates a weighted representation of all transformer layer outputs and
    performs pooling over the individual pieces of each Doc token, returning their
    corresponding representations.

    Requires its upstream Transformer component to return all layer outputs from
    its model.
    """

    name = "scalar_weighting_listener"

    def __init__(
        self,
        upstream_name: str,
        weighting: ScalarWeightModelT,
        pooling: PoolingModelT,
        width: int,
        grad_factor: float,
    ) -> None:
        """
        upstream_name (str):
            A string to identify the 'upstream' Transformer component
            to communicate with. The upstream name should either be the wildcard
            string '*', or the name of the Transformer component.

            In almost all cases, the wildcard string will suffice as there'll only be one
            upstream Transformer component. But in certain situations, e.g: you have disjoint
            datasets for certain tasks, or you'd like to use a pre-trained pipeline but a
            downstream task requires its own token representations, you could end up with
            more than one Transformer component in the pipeline.
        weighting (Model):
            Model that is used to perform the weighting of the different layer outputs.
        pooling (Model):
            Model that is used to perform pooling over the piece representations.
        width (int):
            The width of the vectors produced by the upstream transformer component.
        grad_factor (float):
            Factor to multiply gradients with.
        """
        Model.__init__(
            self,
            name=self.name,
            forward=scalar_weighting_listener_forward,
            dims={"nO": width},
            layers=[weighting, with_ragged_last_layer(pooling)],
            attrs={
                "grad_factor": grad_factor,
            },
        )
        self.upstream_name = upstream_name
        self._batch_id = None
        self._outputs = None
        self._backprop = None


def scalar_weighting_listener_forward(
    model: ScalarWeightingListener, docs: Iterable[Doc], is_train: bool
) -> Tuple[List[Floats2d], Callable[[Any], Any]]:
    weighting: ScalarWeightModelT = model.layers[0]
    pooling: WithRaggedLastLayerModelT = model.layers[1]
    grad_factor: float = model.attrs["grad_factor"]

    if is_train:
        assert model._outputs is not None
        if model._outputs.last_layer_only:
            raise ValueError(
                Errors.E012.format(listener_name="ScalarWeightingListener")
            )

        model.verify_inputs(docs)

        Y_weighting: ScalarWeightOutT = []
        weighting_inputs = model._outputs.all_outputs
        outputs_to_backprop = tuple(i for i in range(model._outputs.num_outputs))

        Y_weighting, backprop_weighting = weighting(weighting_inputs, is_train)
        Y, backprop_pooling = pooling(Y_weighting, is_train)

        def backprop(dYs):
            dX_pooling = backprop_pooling(dYs)
            dX_weighting = backprop_weighting(dX_pooling)

            if grad_factor != 1.0:
                for dx_inner in dX_weighting:
                    for dx in dx_inner:
                        dx.data *= grad_factor

            dX = model._backprop(dX_weighting, outputs_to_backprop=outputs_to_backprop)
            model._batch_id = None
            model._outputs = None
            model._backprop = None
            return dX

        return Y, backprop
    else:
        width = model.get_dim("nO")

        no_trf_data = [doc._.trf_data is None for doc in docs]
        if any(no_trf_data):
            assert all(no_trf_data)
            return [model.ops.alloc2f(len(doc), width) for doc in docs], lambda dY: []

        if any(doc._.trf_data.last_layer_only for doc in docs):
            raise ValueError(
                Errors.E012.format(listener_name="ScalarWeightingListener")
            )

        Y_weigthing = weighting.predict([doc._.trf_data.all_outputs for doc in docs])
        Y = pooling.predict(Y_weigthing)

        return Y, lambda dX: []



from dataclasses import dataclass

from ..bert import BertConfig


@dataclass
class RobertaConfig(BertConfig):
    def __init__(
        self,
        *args,
        layer_norm_eps=1e-05,
        max_position_embeddings=514,
        padding_idx=1,
        type_vocab_size=1,
        vocab_size=50265,
        **kwargs
    ):
        super(RobertaConfig, self).__init__(
            *args,
            layer_norm_eps=layer_norm_eps,
            max_position_embeddings=max_position_embeddings,
            padding_idx=padding_idx,
            type_vocab_size=type_vocab_size,
            vocab_size=vocab_size,
            **kwargs
        )



from typing import List
from cutlery import SentencePieceProcessor
import pytest
from thinc.api import Ragged, chain, Model, get_current_ops, NumpyOps
from thinc.util import convert_recursive, is_cupy_array

from curated_transformers._compat import has_hf_transformers, transformers
from curated_transformers.tokenization.sentencepiece_encoder import (
    build_sentencepiece_encoder_v1,
    build_xlmr_sentencepiece_encoder_v1,
)
from curated_transformers.tokenization.sentencepiece_adapters import (
    build_xlmr_adapter,
)
from curated_transformers.tokenization.wordpiece_encoder import (
    build_wordpiece_encoder_v1,
)
from curated_transformers.tokenization.hf_loader import build_hf_piece_encoder_loader_v1
from curated_transformers.models.output import TransformerModelOutput
from curated_transformers.models.remove_eos_bos import remove_bos_eos


@pytest.fixture
def toy_model(test_dir):
    return SentencePieceProcessor.from_file(str(test_dir / "toy.model"))


@pytest.fixture
def toy_encoder(toy_model):
    encoder = build_xlmr_sentencepiece_encoder_v1()
    encoder.get_ref("encoder").attrs["sentencepiece_processor"] = toy_model
    return encoder


@pytest.fixture
def empty_encoder():
    return build_xlmr_sentencepiece_encoder_v1()


def _mock_transformer() -> Model[List[Ragged], TransformerModelOutput]:
    def forward(model: Model, X: List[Ragged], is_train: bool):
        return (
            TransformerModelOutput(outputs=[[x] for x in X], last_layer_only=True),
            lambda x: x,
        )

    return Model("mock_transformer", forward)


def test_sentencepiece_encoder(toy_encoder, sample_docs):
    encoding = toy_encoder.predict(sample_docs)
    _check_encoder(encoding)


def test_serialize(toy_encoder, empty_encoder, sample_docs):
    encoder_bytes = toy_encoder.to_bytes()
    encoder2 = empty_encoder
    encoder2.from_bytes(encoder_bytes)
    encoding = encoder2.predict(sample_docs)
    _check_encoder(encoding)


def _compare_model_hf_output(ops, Y, Y_hf):
    numpy = NumpyOps().xp
    # Get inputs, removing BOS/EOS from every token.
    Y_hf = [numpy.array(e[1:-1]) for e in Y_hf["input_ids"]]

    # Since cupy balks at comparing sequences that mix CPU lists and GPU arrays,
    # we'll need to convert them to numpy (CPU) arrays first.
    Y = convert_recursive(
        is_cupy_array, lambda a: a.get(), ops.unflatten(Y.dataXd, Y.lengths)
    )
    numpy.testing.assert_equal(Y, Y_hf)


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_sentencepiece_encoder_against_hf(sample_docs):
    ops = get_current_ops()

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("xlm-roberta-base")
    encoder = build_sentencepiece_encoder_v1()
    encoder.init = build_hf_piece_encoder_loader_v1(name="xlm-roberta-base")
    encoder.initialize()
    model = chain(encoder, build_xlmr_adapter(), _mock_transformer(), remove_bos_eos())

    encoding = model.predict(sample_docs)
    hf_encoding = hf_tokenizer([token.text for token in sample_docs[0]])
    _compare_model_hf_output(ops, encoding.all_outputs[0][0], hf_encoding)

    hf_encoding = hf_tokenizer([token.text for token in sample_docs[1]])
    _compare_model_hf_output(ops, encoding.all_outputs[1][0], hf_encoding)


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_wordpiece_encoder_against_hf(sample_docs):
    ops = get_current_ops()
    encoder = build_wordpiece_encoder_v1()
    encoder.init = build_hf_piece_encoder_loader_v1(name="bert-base-cased")
    encoder.initialize()
    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("bert-base-cased")
    model = chain(encoder, _mock_transformer(), remove_bos_eos())

    encoding = model.predict(sample_docs)
    hf_encoding = hf_tokenizer([token.text for token in sample_docs[0]])
    _compare_model_hf_output(ops, encoding.all_outputs[0][0], hf_encoding)

    hf_encoding = hf_tokenizer([token.text for token in sample_docs[1]])
    _compare_model_hf_output(ops, encoding.all_outputs[1][0], hf_encoding)


def _check_encoder(encoding):
    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops = get_current_ops()
    ops.xp.testing.assert_array_equal(
        encoding[0].lengths, [1, 1, 1, 1, 1, 1, 1, 6, 2, 1]
    )
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd,
        [0, 9, 466, 11, 948, 42, 11, 171, 169, 111, 29, 21, 144, 8, 5, 2],
    )

    assert isinstance(encoding[1], Ragged)
    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 2, 1, 1, 1, 4, 3, 2, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd,
        [0, 484, 547, 113, 172, 568, 63, 21, 46, 3, 85, 116, 28, 8, 5, 2],
    )



from dataclasses import dataclass


# Defaults taken from syntaxdot
# https://github.com/tensordot/syntaxdot/blob/22bd3d43ed2d7fcbef8a6217b01684194fae713f/syntaxdot-transformers/src/models/bert/config.rs#L25


@dataclass
class BertEmbeddingConfig:
    embedding_width: int
    vocab_size: int
    type_vocab_size: int
    max_position_embeddings: int
    layer_norm_eps: float
    dropout_prob: float

    def __init__(
        self,
        *,
        embedding_width: int = 768,
        vocab_size: int = 30000,
        type_vocab_size: int = 2,
        max_position_embeddings: int = 512,
        layer_norm_eps: float = 1e-12,
        dropout_prob: float = 0.1,
    ) -> None:
        self.embedding_width = embedding_width
        self.vocab_size = vocab_size
        self.type_vocab_size = type_vocab_size
        self.max_position_embeddings = max_position_embeddings
        self.layer_norm_eps = layer_norm_eps
        self.dropout_prob = dropout_prob


@dataclass
class BertAttentionConfig:
    hidden_width: int
    num_attention_heads: int
    dropout_prob: float

    def __init__(
        self,
        *,
        hidden_width: int = 768,
        num_attention_heads: int = 12,
        dropout_prob: float = 0.1,
    ) -> None:
        self.hidden_width = hidden_width
        self.num_attention_heads = num_attention_heads
        self.dropout_prob = dropout_prob


@dataclass
class BertLayerConfig:
    hidden_width: int
    intermediate_width: int
    num_hidden_layers: int
    hidden_act: str
    layer_norm_eps: float
    dropout_prob: float

    def __init__(
        self,
        *,
        hidden_width: int = 768,
        intermediate_width: int = 3072,
        num_hidden_layers: int = 12,
        hidden_act: str = "gelu",
        layer_norm_eps: float = 1e-12,
        dropout_prob: float = 1.0,
    ) -> None:
        self.hidden_width = hidden_width
        self.intermediate_width = intermediate_width
        self.num_hidden_layers = num_hidden_layers
        self.hidden_act = hidden_act
        self.layer_norm_eps = layer_norm_eps
        self.dropout_prob = dropout_prob


@dataclass
class BertConfig:
    embedding: BertEmbeddingConfig
    attention: BertAttentionConfig
    layer: BertLayerConfig
    model_max_length: int
    padding_idx: int

    def __init__(
        self,
        *,
        embedding_width: int = 768,
        hidden_width: int = 768,
        intermediate_width: int = 3072,
        num_attention_heads: int = 12,
        num_hidden_layers: int = 12,
        attention_probs_dropout_prob: float = 0.1,
        hidden_dropout_prob: float = 0.1,
        hidden_act: str = "gelu",
        vocab_size: int = 30000,
        type_vocab_size: int = 2,
        max_position_embeddings: int = 512,
        model_max_length: int = 512,
        layer_norm_eps: float = 1e-12,
        padding_idx: int = 0,
    ):
        self.embedding = BertEmbeddingConfig(
            embedding_width=embedding_width,
            vocab_size=vocab_size,
            type_vocab_size=type_vocab_size,
            max_position_embeddings=max_position_embeddings,
            layer_norm_eps=layer_norm_eps,
            dropout_prob=hidden_dropout_prob,
        )
        self.attention = BertAttentionConfig(
            hidden_width=hidden_width,
            num_attention_heads=num_attention_heads,
            dropout_prob=attention_probs_dropout_prob,
        )
        self.layer = BertLayerConfig(
            hidden_width=hidden_width,
            intermediate_width=intermediate_width,
            num_hidden_layers=num_hidden_layers,
            hidden_act=hidden_act,
            layer_norm_eps=layer_norm_eps,
            dropout_prob=hidden_dropout_prob,
        )
        self.model_max_length = model_max_length
        self.padding_idx = padding_idx



import torch
from torch import Tensor
from torch.nn import Module

from ...errors import Errors


# From syntaxdot:
# https://github.com/tensordot/syntaxdot/blob/22bd3d43ed2d7fcbef8a6217b01684194fae713f/syntaxdot-transformers/src/scalar_weighting.rs#L62
class ScalarWeight(Module):
    def __init__(self, *, num_layers: int, dropout_prob: float = 0.1):
        super().__init__()

        self.layer_weights = torch.nn.parameter.Parameter(torch.zeros(num_layers))
        self.scale = torch.nn.parameter.Parameter(torch.tensor((1.0,)))
        self.dropout_prob = dropout_prob

    def forward(
        self,
        layer_outputs: Tensor,
    ) -> Tensor:
        """
        Shapes:
            layer_outputs - (batch_size, seq_len, num_layers, width)

        Returns a weighted tensor of the input with shape (batch_size, seq_len, width).
        """
        if layer_outputs.shape[2] != self.layer_weights.shape[0]:
            raise ValueError(
                Errors.E008.format(
                    num_layers_scalar_weight=self.layer_weights.shape[0],
                    num_layers_transformer=layer_outputs.shape[1],
                )
            )

        if self.training:
            dropout_mask = torch.full_like(
                self.layer_weights, 1.0 - self.dropout_prob
            ).bernoulli()
            softmask_mask = (1.0 - dropout_mask) * -10000.0
            layer_weights = self.layer_weights + softmask_mask
        else:
            layer_weights = self.layer_weights

        # Convert the layer weights into a probability distribution and
        # expand dimensions to get shape [1, 1, n_layers, 1].
        layer_weights = (
            layer_weights.softmax(dim=-1).unsqueeze(0).unsqueeze(0).unsqueeze(-1)
        )
        weighted_layers = layer_outputs * layer_weights

        return weighted_layers.sum(dim=-2, keepdim=False) * self.scale



from thinc.api import Ragged, get_current_ops
import torch

from curated_transformers.models.scalar_weight import build_scalar_weight_v1


def test_scalar_weight_model():
    ops = get_current_ops()
    model = build_scalar_weight_v1(num_layers=2, dropout_prob=0.0)

    with torch.no_grad():
        model.shims[0]._model.layer_weights[0] = 1
        model.shims[0]._model.layer_weights[1] = 1
        model.shims[0]._model.layer_weights[2] = 1

    zeros = ops.alloc2f(15, 2)
    ones = ops.alloc2f(15, 2) + 1

    lens = ops.asarray1i([1, 2, 3, 4, 5])
    X = [
        [
            Ragged(ones.copy(), lens.copy()),
            Ragged(zeros.copy(), lens.copy()),
            Ragged(ones.copy(), lens.copy()),
        ]
    ]
    Y = ops.alloc2f(15, 2) + (1.0 / 3.0) * 2

    Yh, backprop = model(X, is_train=True)
    assert len(Yh) == 1
    ops.xp.testing.assert_array_equal(
        Yh[0].dataXd,
        Y,
    )
    ops.xp.testing.assert_array_equal(
        Yh[0].lengths,
        lens,
    )

    dX = backprop(
        [Ragged(ones.copy(), lengths=lens.copy())],
    )
    dX_expected = ops.alloc2f(15, 2) + (1.0 / 3.0)

    assert len(dX) == 1
    assert len(dX[0]) == 3
    for dX_layer in dX[0]:
        ops.xp.testing.assert_array_equal(
            dX_layer.dataXd,
            dX_expected,
        )
        ops.xp.testing.assert_array_equal(
            dX_layer.lengths,
            lens,
        )



import pytest
import numpy
from torch.nn import Linear
from thinc.layers import (
    PyTorchWrapper_v2,
    TorchScriptWrapper_v1,
    pytorch_to_torchscript_wrapper,
)
from thinc.api import get_array_module


@pytest.mark.parametrize("nN,nI,nO", [(2, 3, 4)])
def test_pytorch_script(nN, nI, nO):
    model = PyTorchWrapper_v2(Linear(nI, nO)).initialize()

    script_model = pytorch_to_torchscript_wrapper(model)

    X = numpy.random.randn(nN, nI).astype("f")
    Y = model.predict(X)
    Y_script = script_model.predict(X)
    xp = get_array_module(Y)
    xp.testing.assert_array_equal(Y, Y_script)

    serialized = script_model.to_bytes()
    script_model2 = TorchScriptWrapper_v1()
    script_model2.from_bytes(serialized)

    xp.testing.assert_array_equal(Y, script_model2.predict(X))
